import numpy as np
import astropy.constants as const
import astropy.units as u
from astropy.table import Table
from pkg_resources import resource_filename
from abc import ABCMeta, abstractmethod

k_B = const.k_B.to("eV / kK").value
c3 = (4. * np.pi * const.sigma_sb.to("erg s-1 Rsun-2 kK-4").value) ** -0.5 / 1000.  # Rsun --> kiloRsun
c4 = 1. / (4. * np.pi * u.Mpc.to(u.m) ** 2.)


def format_unit(unit):
    """
    Use LaTeX to format a physical unit, which may consist of an order of magnitude times a base unit

    Parameters
    ----------
    unit : astropy.units.Unit, astropy.units.Quantity
        The unit to format

    Returns
    -------
    unit_str : str
        Formatted unit
    """
    if isinstance(unit, u.Quantity):
        value = np.log10(unit.value)
        unit = unit.unit
        if value % 1.:
            unit_str = '$10^{{{value:.1f}}}$ {unit:latex_inline}'
        else:
            unit_str = '$10^{{{value:.0f}}}$ {unit:latex_inline}'
    else:
        value = None
        unit_str = '{unit:latex_inline}'
    return unit_str.format(value=value, unit=unit)


class Model:
    """
    An analytical model, defined by a function and its parameters

    Parameters
    ----------
    func : function
        A function that defines the analytical model
    input_names : iterable
        A list of parameter names
    units : iterable
        A list of units (:class:`astropy.unit.Unit`) for each parameter

    Attributes
    ----------
    func : function
        The function that defines the analytical model
    input_names : iterable
        A list of the parameter names
    units : iterable
        A list of the units for each parameter
    nparams : int
        The number of parameters in the model
    axis_labels : list
        Axis labels for each paramter (including name and unit)
    output_quantity : str, optional
        Quantity output by the model: 'lum' (default) or 'flux'
    """
    def __init__(self, func, input_names, units, output_quantity='lum'):
        self.func = func
        self.input_names = input_names
        self.units = units
        self.nparams = len(input_names)
        self.axis_labels = ['${}$ ({})'.format(var, format_unit(unit))
                            if unit is not u.dimensionless_unscaled else '${}$'.format(var)
                            for var, unit in zip(input_names, units)]
        self.output_quantity = output_quantity

    def __call__(self, *args, **kwargs):
        return self.func(*args[:self.nparams+2], **kwargs)  # +2 for times and filters


def shock_cooling_temperature_radius(t_in, v_s, M_env, f_rho_M, R, t_exp=0., kappa=1., n=1.5, RW=False):
    """
    The shock cooling model of Sapir & Waxman (https://doi.org/10.3847/1538-4357/aa64df).

    This version of the model is written in terms of physical parameters :math:`v_s, M_\\mathrm{env}, f_ρ M, R`:

    :math:`T(t) = \\frac{T_\\mathrm{col}}{T_\\mathrm{ph}} T_0 \\left(\\frac{v_s^2 t^2}{f_ρ M κ}\\right)^{ε_1}
    \\frac{R^{1/4}}{κ^{1/4}} t^{-1/2}` (Eq. 23)

    :math:`L(t) = A \\exp\\left[-\\left(\\frac{a t}{t_\\mathrm{tr}}\\right)^α\\right]
    L_0 \\left(\\frac{v_s t^2}{f_ρ M κ}\\right)^{-ε_2} \\frac{v_s^2 R}{κ}` (Eq. 18-19)

    :math:`t_\\mathrm{tr} = (19.5\\,\\mathrm{d}) \\left(\\frac{κ * M_\\mathrm{env}}{v_s} \\right)^{1/2}` (Eq. 20)

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    v_s : float, array-like
        The shock speed in :math:`10^{8.5}` cm/s
    M_env : float, array-like
        The envelope mass in solar masses
    f_rho_M : float, array-like
        The product :math:`f_ρ M`, where ":math:`f_ρ` is a numerical factor of order unity that depends on the inner
        envelope structure" and :math:`M` is the ejecta mass in solar masses
    R : float, array-like
        The progenitor radius in :math:`10^{13}` cm
    t_exp : float, array-like
        The explosion epoch
    kappa : float, array-like
        The ejecta opacity in units of the electron scattering opacity (0.34 cm^2/g)
    n : float, array-like
        The polytropic index of the progenitor. Must be either 1.5 or 3.
    RW : bool, optional
        Reduce the model to the simpler form of Rabinak & Waxman (https://doi.org/10.1088/0004-637X/728/1/63)

    Returns
    -------
    T_K : array-like
        The model blackbody temperatures in units of kilokelvins
    R_bb : array-like
        The model blackbody radii in units of 1000 solar radii
    """
    if n == 1.5:
        A = 0.94
        a = 1.67
        alpha = 0.8
        epsilon_1 = 0.027
        epsilon_2 = 0.086
        L_0 = 2.0e42
        T_0 = 1.61
        Tph_to_Tcol = 1.1
    elif n == 3.:
        A = 0.79
        a = 4.57
        alpha = 0.73
        epsilon_1 = 0.016
        epsilon_2 = 0.175
        L_0 = 2.1e42
        T_0 = 1.69
        Tph_to_Tcol = 1.0
    else:
        raise ValueError('n can only be 1.5 or 3')

    if RW:
        a = 0.
        Tph_to_Tcol = 1.2

    t = t_in.reshape(-1, 1) - t_exp
    L_RW = L_0 * (t ** 2 * v_s / (f_rho_M * kappa)) ** -epsilon_2 * v_s ** 2 * R / kappa
    t_tr = 19.5 * (kappa * M_env / v_s) ** 0.5
    L = L_RW * A * np.exp(-(a * t / t_tr) ** alpha)
    T_ph = T_0 * (t ** 2 * v_s ** 2 / (f_rho_M * kappa)) ** epsilon_1 * kappa ** -0.25 * t ** -0.5 * R ** 0.25
    T_col = T_ph * Tph_to_Tcol
    T_K = np.squeeze(T_col) / k_B
    R_bb = c3 * np.squeeze(L) ** 0.5 * T_K ** -2
    return T_K, R_bb


def shock_cooling(t_in, f, v_s, M_env, f_rho_M, R, t_exp=0., kappa=1., n=1.5, RW=False, z=0.):
    """
    The shock cooling model of Sapir & Waxman (https://doi.org/10.3847/1538-4357/aa64df).

    This version of the model is written in terms of physical parameters :math:`v_s, M_\\mathrm{env}, f_ρ M, R`:

    :math:`T(t) = \\frac{T_\\mathrm{col}}{T_\\mathrm{ph}} T_0 \\left(\\frac{v_s^2 t^2}{f_ρ M κ}\\right)^{ε_1}
    \\frac{R^{1/4}}{κ^{1/4}} t^{-1/2}` (Eq. 23)

    :math:`L(t) = A \\exp\\left[-\\left(\\frac{a t}{t_\\mathrm{tr}}\\right)^α\\right]
    L_0 \\left(\\frac{v_s t^2}{f_ρ M κ}\\right)^{-ε_2} \\frac{v_s^2 R}{κ}` (Eq. 18-19)

    :math:`t_\\mathrm{tr} = (19.5\\,\\mathrm{d}) \\left(\\frac{κ * M_\\mathrm{env}}{v_s} \\right)^{1/2}` (Eq. 20)

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    f : lightcurve_fitting.filter.Filter, array-like
        Filters for which to calculate the model
    v_s : float, array-like
        The shock speed in :math:`10^{8.5}` cm/s
    M_env : float, array-like
        The envelope mass in solar masses
    f_rho_M : float, array-like
        The product :math:`f_ρ M`, where ":math:`f_ρ` is a numerical factor of order unity that depends on the inner
        envelope structure" and :math:`M` is the ejecta mass in solar masses
    R : float, array-like
        The progenitor radius in :math:`10^{13}` cm
    t_exp : float, array-like
        The explosion epoch
    kappa : float, array-like
        The ejecta opacity in units of the electron scattering opacity (0.34 cm^2/g)
    n : float, array-like
        The polytropic index of the progenitor. Must be either 1.5 or 3.
    RW : bool, optional
        Reduce the model to the simpler form of Rabinak & Waxman (https://doi.org/10.1088/0004-637X/728/1/63)
    z : float, optional
        The redshift between blackbody source and the observed filters

    Returns
    -------
    y_fit : array-like
        The filtered model light curves
    """
    T_K, R_bb = shock_cooling_temperature_radius(t_in, v_s, M_env, f_rho_M, R, t_exp, kappa, n, RW)
    y_fit = blackbody_to_filters(f, T_K, R_bb, z)
    return y_fit


def t_min(p, kappa=1.):
    """
    The minimum validity time for the :func:`shock_cooling` model

    :math:`t_\\mathrm{min} = (0.2\\,\\mathrm{d}) \\frac{R}{v_s}
    \\max\\left[0.5, \\frac{R^{0.4}}{(f_ρ M κ)^{0.2} v_s^{-0.7}}\\right] + t_\\mathrm{exp}` (Eq. 17)
    """
    v_s = p[0]
    f_rho_M = p[2]
    R = p[3]
    t_exp = p[4] if len(p) > 4 else 0.
    return 0.2 * R / v_s * np.maximum(0.5, R ** 0.4 * (f_rho_M * kappa) ** -0.2 * v_s ** -0.7) + t_exp


def t_max(p, kappa=1.):
    """
    The maximum validity time for the :func:`shock_cooling` model

    :math:`t_\\mathrm{max} = (7.4\\,\\mathrm{d}) \\left(\\frac{R}{κ}\\right)^{0.55} + t_\\mathrm{exp}` (Eq. 24)
    """
    R = p[3]
    t_exp = p[4] if len(p) > 4 else 0.
    return 7.4 * (R / kappa) ** 0.55 + t_exp


ShockCooling = Model(shock_cooling,
                     [
                         'v_\\mathrm{s*}',
                         'M_\\mathrm{env}',
                         'f_\\rho M',
                         'R',
                         't_0'
                     ],
                     [
                         10. ** 8.5 * u.cm / u.s,
                         u.Msun,
                         u.Msun,
                         1e13 * u.cm,
                         u.d
                     ]
                     )
ShockCooling.t_min = t_min
ShockCooling.t_max = t_max


def shock_cooling3(t_in, f, v_s, M_env, f_rho_M, R, dist, ebv=0., t_exp=0., kappa=1., n=1.5, RW=False, z=0.):
    """
    The shock cooling model of Sapir & Waxman (https://doi.org/10.3847/1538-4357/aa64df).

    This version of the model is written in terms of physical parameters :math:`v_s, M_\\mathrm{env}, f_ρ M, R`:

    :math:`T(t) = \\frac{T_\\mathrm{col}}{T_\\mathrm{ph}} T_0 \\left(\\frac{v_s^2 t^2}{f_ρ M κ}\\right)^{ε_1}
    \\frac{R^{1/4}}{κ^{1/4}} t^{-1/2}` (Eq. 23)

    :math:`L(t) = A \\exp\\left[-\\left(\\frac{a t}{t_\\mathrm{tr}}\\right)^α\\right]
    L_0 \\left(\\frac{v_s t^2}{f_ρ M κ}\\right)^{-ε_2} \\frac{v_s^2 R}{κ}` (Eq. 18-19)

    :math:`t_\\mathrm{tr} = (19.5\\,\\mathrm{d}) \\left(\\frac{κ * M_\\mathrm{env}}{v_s} \\right)^{1/2}` (Eq. 20)

    This is the same as :func:`shock_cooling` but with distance and reddening as free parameters.

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    f : lightcurve_fitting.filter.Filter, array-like
        Filters for which to calculate the model
    v_s : float, array-like
        The shock speed in :math:`10^{8.5}` cm/s
    M_env : float, array-like
        The envelope mass in solar masses
    f_rho_M : float, array-like
        The product :math:`f_ρ M`, where ":math:`f_ρ` is a numerical factor of order unity that depends on the inner
        envelope structure" and :math:`M` is the ejecta mass in solar masses
    R : float, array-like
        The progenitor radius in :math:`10^{13}` cm
    dist : float, array-like
        The luminosity distance to the supernova in Mpc
    ebv : float, array-like
        The reddening :math:`E(B-V)` to apply to the blackbody spectrum before integration
    t_exp : float, array-like
        The explosion epoch
    kappa : float, array-like
        The ejecta opacity in units of the electron scattering opacity (0.34 cm^2/g)
    n : float, array-like
        The polytropic index of the progenitor. Must be either 1.5 or 3.
    RW : bool, optional
        Reduce the model to the simpler form of Rabinak & Waxman (https://doi.org/10.1088/0004-637X/728/1/63)
    z : float, optional
        The redshift between blackbody source and the observed filters

    Returns
    -------
    y_fit : array-like
        The filtered model light curves
    """
    T_K, R_bb = shock_cooling_temperature_radius(t_in, v_s, M_env, f_rho_M, R, t_exp, kappa, n, RW)
    lum = blackbody_to_filters(f, T_K, R_bb, z, ebv=ebv)
    flux = c4 * lum / dist ** 2.
    return flux


def t_min3(p, kappa=1.):
    """
    The minimum validity time for the :func:`shock_cooling` model

    :math:`t_\\mathrm{min} = (0.2\\,\\mathrm{d}) \\frac{R}{v_s}
    \\max\\left[0.5, \\frac{R^{0.4}}{(f_ρ M κ)^{0.2} v_s^{-0.7}}\\right] + t_\\mathrm{exp}` (Eq. 17)
    """
    return t_min([p[0], p[1], p[2], p[3], p[6] if len(p) > 6 else 0.], kappa)


def t_max3(p, kappa=1.):
    """
    The maximum validity time for the :func:`shock_cooling` model

    :math:`t_\\mathrm{max} = (7.4\\,\\mathrm{d}) \\left(\\frac{R}{κ}\\right)^{0.55} + t_\\mathrm{exp}` (Eq. 24)
    """
    return t_max([p[0], p[1], p[2], p[3], p[6] if len(p) > 6 else 0.], kappa)


ShockCooling3 = Model(shock_cooling3,
                      [
                          'v_\\mathrm{s*}',
                          'M_\\mathrm{env}',
                          'f_\\rho M',
                          'R',
                          'd_L',
                          'E(B-V)',
                          't_0',
                      ],
                      [
                          10. ** 8.5 * u.cm / u.s,
                          u.Msun,
                          u.Msun,
                          1e13 * u.cm,
                          u.Mpc,
                          u.mag,
                          u.d,
                      ],
                      output_quantity='flux')
ShockCooling3.t_min = t_min3
ShockCooling3.t_max = t_max3


def shock_cooling2(t_in, f, T_1, L_1, t_tr, t_exp=0., n=1.5, RW=False, z=0.):
    """
    The shock cooling model of Sapir & Waxman (https://doi.org/10.3847/1538-4357/aa64df).

    This version of the model is written in terms of scaling parameters :math:`T_1, L_1, t_\\mathrm{tr}`:

    :math:`T(t) = T_1 t^{2 ε_1 - 0.5}` and
    :math:`L(t) = L_1 \\exp\\left[-\\left(\\frac{a t}{t_\\mathrm{tr}}\\right)^α\\right] t^{-2 ε_2}`

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    f : lightcurve_fitting.filter.Filter, array-like
        Filters for which to calculate the model
    T_1 : float, array-like
        The blackbody temperature 1 day after explosion in kilokelvins
    L_1 : float, array-like
        The approximate blackbody luminosity 1 day after explosion in :math:`10^{42}` erg/s
    t_tr : float, array-like
        The time at which the envelope becomes transparent in rest-frame days
    t_exp : float, array-like
        The explosion epoch
    n : float, array-like
        The polytropic index of the progenitor. Must be either 1.5 or 3.
    RW : bool, optional
        Reduce the model to the simpler form of Rabinak & Waxman (https://doi.org/10.1088/0004-637X/728/1/63)
    z : float, optional
        The redshift between blackbody source and the observed filters

    Returns
    -------
    y_fit : array-like
        The filtered model light curves
    """
    if n == 1.5:
        a = 1.67
        alpha = 0.8
        epsilon_1 = 0.027
        epsilon_2 = 0.086
    elif n == 3.:
        a = 4.57
        alpha = 0.73
        epsilon_1 = 0.016
        epsilon_2 = 0.175
    else:
        raise ValueError('n can only be 1.5 or 3')

    if RW:
        a = 0.

    t = t_in.reshape(-1, 1) - t_exp

    epsilon_T = 2 * epsilon_1 - 0.5
    epsilon_L = -2 * epsilon_2

    T_K = np.squeeze(T_1 * t ** epsilon_T)
    L = np.squeeze(L_1 * np.exp(-(a * t / t_tr) ** alpha) * t ** epsilon_L) * 1e42
    R_bb = c3 * L ** 0.5 * T_K ** -2

    y_fit = blackbody_to_filters(f, T_K, R_bb, z)

    return y_fit


def t_min2(*args):
    """
    The maximum validity time for the :func:`shock_cooling2` model

    Raises
    ------
    NotImplementedError
    """
    raise NotImplementedError('t_min cannot be translated to these parameters')


def t_max2(p, n=1.5):
    """
    The maximum validity time for the :func:`shock_cooling2` model

    :math:`t_\\mathrm{max} = \\left(\\frac{8.12\\,\\mathrm{kK}}{T_1}\\right)^{1/(2 ε_1 - 0.5)} + t_\\mathrm{exp}`
    """
    T_1 = p[0]
    t_exp = p[3] if len(p) > 3 else 0.

    if n == 1.5:
        epsilon_1 = 0.027
    elif n == 3.:
        epsilon_1 = 0.016
    else:
        raise ValueError('n can only be 1.5 or 3')

    epsilon_T = 2 * epsilon_1 - 0.5

    return (8.12 / T_1) ** (epsilon_T ** -1) + t_exp


ShockCooling2 = Model(shock_cooling2,
                      [
                          'T_1',
                          'L_1',
                          't_\\mathrm{tr}',
                          't_0'
                      ],
                      [
                          u.kK,
                          1e42 * u.erg / u.s,
                          u.d,
                          u.d
                      ]
                      )
ShockCooling2.t_min = t_min2
ShockCooling2.t_max = t_max2

sifto_filename = resource_filename('lightcurve_fitting', 'models/sifto.dat')
sifto = Table.read(sifto_filename, format='ascii')
sifto['x'] = sifto['r']  # assume DLT40 = r for now


def scale_sifto(sn_lc):
    """
    Scale the SiFTO model to match your supernova's luminosity and colors

    Parameters
    ----------
    sn_lc : lightcurve_fitting.lightcurve.LC
        Your supernova's light curve
    """
    for filt in set(sn_lc['filter']):
        if filt.char not in sifto.colnames:
            raise Exception('No SiFTO template for filter ' + filt.char)
        lc_filt = sn_lc.where(filter=filt)
        sifto[filt.char] *= np.max(lc_filt['lum']) / np.max(sifto[filt.char])


def companion_shocking_temperature_radius(t_in, t_exp, a13, Mc_v9_7, kappa=1.):
    """
    The companion shocking model of Kasen (https://doi.org/10.1088/0004-637X/708/2/1025)

    As written by Hosseinzadeh et al. (https://doi.org/10.3847/2041-8213/aa8402), the shock component is defined by:

    :math:`R_\\mathrm{phot}(t) = (2700\\,R_\\odot) (M_c v_9^7)^{1/9} κ^{1/9} t^{7/9}` (Eq. 1)

    :math:`T_\\mathrm{eff}(t) = (25\\,\\mathrm{kK}) a_{13}^{1/4} (M_c v_9^7)^{1/144} κ^{-35/144} t^{37/72}` (Eq. 2)

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    t_exp : float, array-like
        The explosion epoch
    a13 : float, array-like
        The binary separation in :math:`10^{13}` cm
    Mc_v9_7 : float, array-like
        The product :math:`M_c v_9^7`, where :math:`M_c` is the ejecta mass in Chandrasekhar masses and :math:`v_9` is
        the ejecta velocity in units of :math:`10^9` cm/s
    kappa : float, array-like
        The ejecta opacity in units of the electron scattering opacity (0.34 cm^2/g)

    Returns
    -------
    T_kasen : array-like
        The model blackbody temperatures in units of kilokelvins
    R_kasen : array-like
        The model blackbody radii in units of 1000 solar radii
    """
    t_wrt_exp = t_in.reshape(-1, 1) - t_exp
    T_kasen = np.squeeze(25. * (a13 ** 36 * Mc_v9_7 * kappa ** -35 * t_wrt_exp ** -74) ** (1 / 144.))  # kK
    R_kasen = np.squeeze(2.7 * (kappa * Mc_v9_7 * t_wrt_exp ** 7) ** (1 / 9.))  # kiloRsun
    return T_kasen, R_kasen


def companion_shocking(t_in, f, t_exp, a13, Mc_v9_7, kappa=1., z=0., cutoff_freq=np.inf):
    """
    The companion shocking model of Kasen (https://doi.org/10.1088/0004-637X/708/2/1025)

    As written by Hosseinzadeh et al. (https://doi.org/10.3847/2041-8213/aa8402), the shock component is defined by:

    :math:`R_\\mathrm{phot}(t) = (2700\\,R_\\odot) (M_c v_9^7)^{1/9} κ^{1/9} t^{7/9}` (Eq. 1)

    :math:`T_\\mathrm{eff}(t) = (25\\,\\mathrm{kK}) a_{13}^{1/4} (M_c v_9^7)^{1/144} κ^{-35/144} t^{37/72}` (Eq. 2)

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    f : lightcurve_fitting.filter.Filter, array-like
        Filters for which to calculate the model
    t_exp : float, array-like
        The explosion epoch
    a13 : float, array-like
        The binary separation in :math:`10^{13}` cm
    Mc_v9_7 : float, array-like
        The product :math:`M_c v_9^7`, where :math:`M_c` is the ejecta mass in Chandrasekhar masses and :math:`v_9` is
        the ejecta velocity in units of :math:`10^9` cm/s
    kappa : float, array-like
        The ejecta opacity in units of the electron scattering opacity (0.34 cm^2/g)
    z : float, optional
        The redshift between blackbody source and the observed filters
    cutoff_freq : float, optional
        Cutoff frequency (in terahertz) for a modified blackbody spectrum (see https://doi.org/10.3847/1538-4357/aa9334)

    Returns
    -------
    y_fit : array-like
        The filtered model light curves
    """
    T_kasen, R_kasen = companion_shocking_temperature_radius(t_in, t_exp, a13, Mc_v9_7, kappa)
    Lnu_kasen = blackbody_to_filters(f, T_kasen, R_kasen, z, cutoff_freq)
    return Lnu_kasen


def stretched_sifto(t_in, f, t_peak, stretch, dtU=None, dti=None):
    """
    The SiFTO SN Ia model (https://doi.org/10.1086/588518), offset and stretched by the input parameters

    The SiFTO model is currently only available in the UBVgri filters. We assume DLT40 can be modeled as r.

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    f : lightcurve_fitting.filter.Filter, array-like
        Filters for which to calculate the model
    t_peak : float, array-like
        The epoch of maximum light for the SiFTO model
    stretch : float, array-like
        The stretch for the SiFTO model
    dtU, dti : float, array-like
        Time offsets for the U- and i-band models relative to the other bands

    Returns
    -------
    y_fit : array-like
        The filtered model light curves
    """
    dt_peak = {'U': np.zeros_like(stretch) if dtU is None else dtU, 'i': np.zeros_like(stretch) if dti is None else dti}
    t_wrt_peak = np.squeeze(t_in.reshape(-1, 1) - t_peak)
    if t_wrt_peak.ndim <= 1 and len(t_wrt_peak) == len(f):  # pointwise
        Lnu_sifto = np.array([np.interp(t - dt_peak.get(filt.char, 0.), sifto['Epoch'] * stretch, sifto[filt.char])
                              for t, filt in zip(t_wrt_peak, f)])
    elif t_wrt_peak.ndim <= 1:
        Lnu_sifto = np.array([np.interp(t_wrt_peak - dt_peak.get(filt.char, 0.), sifto['Epoch'] * stretch, sifto[filt.char])
                              for filt in f])
    else:
        Lnu_sifto = np.array([np.array([np.interp(t - dt, sifto['Epoch'] * s, sifto[filt.char])
                                        for t, s, dt in zip(t_wrt_peak.T, stretch, dt_peak.get(filt.char, np.zeros_like(stretch)))]).T
                              for filt in f])
    return Lnu_sifto


def companion_shocking_plus_sifto(t_in, f, t_exp, a13, Mc_v9_7, t_peak, stretch, rr=1., ri=1., rU=1., kappa=1., z=0.):
    """
    The companion shocking model of Kasen (https://doi.org/10.1088/0004-637X/708/2/1025) plus the SiFTO SN Ia model

    As written by Hosseinzadeh et al. (https://doi.org/10.3847/2041-8213/aa8402), the shock component is defined by:

    :math:`R_\\mathrm{phot}(t) = (2700\\,R_\\odot) (M_c v_9^7)^{1/9} κ^{1/9} t^{7/9}` (Eq. 1)

    :math:`T_\\mathrm{eff}(t) = (25\\,\\mathrm{kK}) a_{13}^{1/4} (M_c v_9^7)^{1/144} κ^{-35/144} t^{37/72}` (Eq. 2)

    The SiFTO model (https://doi.org/10.1086/588518) is currently only available in the UBVgri filters.

    This version of the model includes factors on the r and i SiFTO models, and a factor on the U shock component.

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    f : lightcurve_fitting.filter.Filter, array-like
        Filters for which to calculate the model
    t_exp : float, array-like
        The explosion epoch
    a13 : float, array-like
        The binary separation in :math:`10^{13}` cm
    Mc_v9_7 : float, array-like
        The product :math:`M_c v_9^7`, where :math:`M_c` is the ejecta mass in Chandrasekhar masses and :math:`v_9` is
        the ejecta velocity in units of :math:`10^9` cm/s
    t_peak : float, array-like
        The epoch of maximum light for the SiFTO model
    stretch : float, array-like
        The stretch for the SiFTO model
    rr : float, array-like
        A scale factor for the r-band SiFTO model
    ri : float, array-like
        A scale factor for the i-band SiFTO model
    rU : float, aray-like
        A scale factor for the U-band of the shock component
    kappa : float, array-like
        The ejecta opacity in units of the electron scattering opacity (0.34 cm^2/g)
    z : float, optional
        The redshift between blackbody source and the observed filters

    Returns
    -------
    y_fit : array-like
        The filtered model light curves
    """
    Lnu_kasen = companion_shocking(t_in, f, t_exp, a13, Mc_v9_7, kappa, z)
    Lnu_sifto = stretched_sifto(t_in, f, t_peak, stretch)

    sifto_factors = {'r': rr, 'i': ri}
    kasen_factors = {'U': rU}
    y_fit = np.array([L1 * kasen_factors.get(filt.char, 1.) + L2 * sifto_factors.get(filt.char, 1.)
                      for L1, L2, filt in zip(Lnu_kasen, Lnu_sifto, f)])

    return y_fit


M_chandra = u.def_unit('M_chandra', 1.4 * u.Msun, format={'latex': 'M_\\mathrm{Ch}'})
CompanionShocking = Model(companion_shocking_plus_sifto,
                          [
                              't_0',
                              'a',
                              'M v^7',
                              't_\\mathrm{max}',
                              's',
                              'r_r',
                              'r_i',
                              'r_U'
                          ],
                          [
                              u.d,
                              10. ** 13. * u.cm,
                              M_chandra * (1e9 * u.cm / u.s) ** 7,
                              u.d,
                              u.dimensionless_unscaled,
                              u.dimensionless_unscaled,
                              u.dimensionless_unscaled,
                              u.dimensionless_unscaled
                          ]
                          )


def companion_shocking_plus_sifto2(t_in, f, t_exp, a13, Mc_v9_7, t_peak, stretch, dtU=0., dti=0., kappa=1., z=0.):
    """
    The companion shocking model of Kasen (https://doi.org/10.1088/0004-637X/708/2/1025) plus the SiFTO SN Ia model

    As written by Hosseinzadeh et al. (https://doi.org/10.3847/2041-8213/aa8402), the shock component is defined by:

    :math:`R_\\mathrm{phot}(t) = (2700\\,R_\\odot) (M_c v_9^7)^{1/9} κ^{1/9} t^{7/9}` (Eq. 1)

    :math:`T_\\mathrm{eff}(t) = (25\\,\\mathrm{kK}) a_{13}^{1/4} (M_c v_9^7)^{1/144} κ^{-35/144} t^{37/72}` (Eq. 2)

    The SiFTO model (https://doi.org/10.1086/588518) is currently only available in the UBVgri filters.

    This version of the model includes time offsets for the U and i SiFTO models.

    Parameters
    ----------
    t_in : float, array-like
        Time in days
    f : lightcurve_fitting.filter.Filter, array-like
        Filters for which to calculate the model
    t_exp : float, array-like
        The explosion epoch
    a13 : float, array-like
        The binary separation in :math:`10^{13}` cm
    Mc_v9_7 : float, array-like
        The product :math:`M_c v_9^7`, where :math:`M_c` is the ejecta mass in Chandrasekhar masses and :math:`v_9` is
        the ejecta velocity in units of :math:`10^9` cm/s
    t_peak : float, array-like
        The epoch of maximum light for the SiFTO model
    stretch : float, array-like
        The stretch for the SiFTO model
    dtU, dti : float, array-like
        Time offsets for the U- and i-band SiFTO models relative to the other bands
    kappa : float, array-like
        The ejecta opacity in units of the electron scattering opacity (0.34 cm^2/g)
    z : float, optional
        The redshift between blackbody source and the observed filters

    Returns
    -------
    y_fit : array-like
        The filtered model light curves
    """
    Lnu_kasen = companion_shocking(t_in, f, t_exp, a13, Mc_v9_7, kappa, z)
    Lnu_sifto = stretched_sifto(t_in, f, t_peak, stretch, dtU, dti)

    kasen_factors = {'U': 1.}
    y_fit = np.array([L1 * kasen_factors.get(filt.char, 1.) + L2 for L1, L2, filt in zip(Lnu_kasen, Lnu_sifto, f)])

    return y_fit


CompanionShocking2 = Model(companion_shocking_plus_sifto2,
                          [
                              't_0',
                              'a',
                              'M v^7',
                              't_\\mathrm{max}',
                              's',
                              '\\Delta t_U',
                              '\\Delta t_i',
                          ],
                          [
                              u.d,
                              10. ** 13. * u.cm,
                              M_chandra * (1e9 * u.cm / u.s) ** 7,
                              u.d,
                              u.dimensionless_unscaled,
                              u.d,
                              u.d,
                          ]
                          )


class Prior:
    __metaclass__ = ABCMeta

    def __init__(self, p_min=-np.inf, p_max=np.inf):
        self.p_min = p_min
        self.p_max = p_max

    def __call__(self, p):
        if self.p_min < p < self.p_max:
            return self.logp(p)
        else:
            return -np.inf

    @abstractmethod
    def logp(self, p):
        pass


class UniformPrior(Prior):
    """
    A uniform prior in the parameter, i.e., :math:`\\frac{dP}{dp} \\propto 1`
    """
    def logp(self, p):
        return np.zeros_like(p)


class LogUniformPrior(Prior):
    """
    A uniform prior in the logarithm of the parameter, i.e., :math:`\\frac{dP}{dp} \\propto \\frac{1}{p}`
    """
    def __init__(self, p_min=0., p_max=np.inf):
        if p_min < 0.:
            raise ValueError('a log-uniform prior cannot have negative limits')
        super().__init__(p_min, p_max)

    def logp(self, p):
        return -np.log(p)


class GaussianPrior(Prior):
    """
    A Gaussian prior centered at `mean` with standard deviation `stddev`, i.e.,
    :math:`\\frac{dP}{dp} \\propto \\exp \\left( \\frac{(p - \\mu)^2}{2 \\sigma^2} \\right)`
    """
    def __init__(self, p_min=-np.inf, p_max=np.inf, mean=0., stddev=1.):
        super().__init__(p_min, p_max)
        self.mean = mean
        self.stddev = stddev

    def logp(self, p):
        return -0.5 * ((p - self.mean) / self.stddev) ** 2.


c1 = (const.h / const.k_B).to(u.kK / u.THz).value
c2 = 8 * np.pi ** 2 * (const.h / const.c ** 2).to(u.W / u.Hz / (1000 * u.Rsun) ** 2 / u.THz ** 3).value


def planck_fast(nu, T, R, cutoff_freq=np.inf):
    """
    The Planck spectrum for a blackbody source

    :math:`L_ν = 4 π R^2 \\frac{2 π h ν^3 c^{-2}}{\\exp(h ν / k_B T) - 1}`

    Parameters
    ----------
    nu : float, array-like
        Frequency in terahertz
    T : float, array-like
        Temperature in kilokelvins
    R : float, array-like
        Radius in units of 1000 solar radii
    cutoff_freq : float, optional
        Cutoff frequency (in terahertz) for a modified blackbody spectrum (see https://doi.org/10.3847/1538-4357/aa9334)

    Returns
    -------
    float, array-like
        The spectral luminosity density (:math:`L_ν`) of the source in watts per hertz
    """
    return c2 * np.squeeze(np.multiply.outer(R ** 2, nu ** 3 * np.minimum(1., cutoff_freq / nu))
                           / (np.exp(c1 * np.multiply.outer(T ** -1, nu)) - 1))


def blackbody_to_filters(filters, T, R, z=0., cutoff_freq=np.inf, ebv=0.):
    """
    The average spectral luminosity density (:math:`L_ν`) of a blackbody as observed through one or more filters

    Parameters
    ----------
    filters : lightcurve_fitting.filters.Filter, array-like
        One or more broadband filters
    T : float, array-like
        Blackbody temperatures in kilokelvins
    R : float, array-like
        Blackbody radii in units of 1000 solar radii
    z : float
        Redshift between the blackbody source and the observed filters
    cutoff_freq : float, optional
        Cutoff frequency (in terahertz) for a modified blackbody spectrum (see https://doi.org/10.3847/1538-4357/aa9334)
    ebv : float, array-like, optional
        Selective extinction E(B-V) in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with R_V=3.1.
        Its shape must be broadcastable to T and R. Default: 0.

    Returns
    -------
    y_fit : float, array-like
        The average spectral luminosity density in each filter in watts per hertz
    """
    T = np.array(T)
    R = np.array(R)
    if T.shape != R.shape:
        raise Exception('T & R must have the same shape')
    np.broadcast(T, ebv)  # check if T and ebv are brodcastable, otherwise raise a ValueError
    if T.ndim == 1 and len(T) == len(filters):  # pointwise
        y_fit = np.array([f.synthesize(planck_fast, t, r, cutoff_freq, z=z, ebv=ebv) for f, t, r in zip(filters, T, R)])
    else:
        y_fit = np.array([f.synthesize(planck_fast, T, R, cutoff_freq, z=z, ebv=ebv) for f in filters])
    return y_fit


def planck(nu, T, R, dT=0., dR=0., cov=0.):
    """
    The Planck spectrum for a blackbody source, including propagation of uncertainties

    Parameters
    ----------
    nu : float, array-like
        Frequency in terahertz
    T : float, array-like
        Temperature in kilokelvins
    R : float, array-like
        Radius in units of 1000 solar radii
    dT : float, array-like, optional
        Uncertainty in the temperature in kilokelvins
    dR : float, array-like, optional
        Uncertainty in the radius in units of 1000 solar radii
    cov : float, array-like, optional
        The covariance between the temperature and radius

    Returns
    -------
    Lnu : float, array-like
        The spectral luminosity density (:math:`L_ν`) of the source in watts per hertz
    dLnu : float, array-like
        The uncertainty in the spectral luminosity density in watts per hertz
    """
    Lnu = planck_fast(nu, T, R)
    if not dT and not dR and not cov:
        return Lnu
    dlogLdT = c1 * nu * T ** -2 / (1 - np.exp(-c1 * nu / T))
    dlogLdR = 2. / R
    dLnu = Lnu * (dlogLdT ** 2 * dT ** 2 + dlogLdR ** 2 * dR ** 2 + 2. * dlogLdT * dlogLdR * cov) ** 0.5
    return Lnu, dLnu



from .filters import all_filters, filtdict, extinction_law
from .models import planck_fast, UniformPrior, LogUniformPrior, GaussianPrior
from .lightcurve import LC

from astropy import constants as const
from astropy import units as u

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoLocator
from scipy.optimize import curve_fit, OptimizeWarning
from scipy.stats import gaussian_kde
import emcee
import corner
import os
from pkg_resources import resource_filename
import warnings

for filt in all_filters:
    filt.read_curve()

plt.style.use(resource_filename('lightcurve_fitting', 'serif.mplstyle'))


def pseudo(temp, radius, z, filter0=filtdict['I'], filter1=filtdict['U'], cutoff_freq=np.inf):
    """
    Integrate a blackbody spectrum between two broadband filters to produce a pseudobolometric luminosity

    Parameters
    ----------
    temp : float, array-like
        Blackbody temperatures in kilokelvins
    radius : float, array-like
        Blackbody radii in units of 1000 solar radii
    z : float
        Redshift between the blackbody (rest frame) and the filters (observed frame)
    filter0, filter1 : lightcurve_fitting.filters.Filter, optional
        Filters to integrate between, where ``filter1`` is bluer than ``filter0``. Default: U to I.
    cutoff_freq : float, optional
        Cutoff frequency (in terahertz) for a modified blackbody spectrum (see https://doi.org/10.3847/1538-4357/aa9334)

    Returns
    -------
    L_opt : float, array-like
        Pseudobolometric luminosity in watts
    """
    freq0 = (filter0.freq_eff - filter0.dfreq / 2.).value
    freq1 = (filter1.freq_eff + filter1.dfreq / 2.).value
    x_optical = np.arange(freq0, freq1)
    y_optical = planck_fast(x_optical * (1. + z), temp, radius, cutoff_freq)
    L_opt = np.trapz(y_optical) * 1e12  # dx = 1 THz
    return L_opt


def plot_chain(chain, labels=None):
    """
    Plot the chain histories for an MCMC run

    Parameters
    ----------
    chain : array-like
        Array containing the MCMC chain history, e.g., from :meth:`emcee.EnsembleSampler.chain`
    labels : iterable, optional
        A list of axis labels for each parameter in the chain

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object containing the chain history plots
    """
    ndim = chain.shape[-1]
    fig, ax = plt.subplots(ndim, figsize=(6., 2. * ndim))
    for i in range(ndim):
        ax[i].plot(chain[:, :, i].T, 'k', alpha=0.2)
        if labels:
            ax[i].set_ylabel(labels[i])
    return fig


def spectrum_mcmc(spectrum, epoch1, priors, starting_guesses, z=0., ebv=0., spectrum_kwargs=None, show=False, outpath='.',
                  nwalkers=10, burnin_steps=200, steps=100, save_chains=False, use_sigma=False, sigma_type='relative',
                  labels=None, freq_min=100., freq_max=1000.):
    """
    Fit the given spectral energy distribution to an epoch of photometry using a Markov-chain Monte Carlo routine

    Parameters
    ----------
    spectrum : function
        Function describing the spectrum. The first argument must be frequency in THz, and it must return spectral
        luminosity in watts per hertz. For a blackbody, you can use :func:`.models.planck_fast`.
    epoch1 : lightcurve_fitting.lightcurve.LC
        A single "epoch" of photometry that defines the observed spectral energy distribution
    priors : list
        Prior probability distributions for each model parameter (and sigma, if used). Available priors:
        :class:`.models.UniformPrior` (default), :class:`.models.LogUniformPrior`, :class:`.models.GaussianPrior`
    starting_guesses : array-like
        Initial guesses for each input parameter to ``spectrum``. Must have shape (nwalkers, nparameters).
    z : float, optional
        Redshift between the emission source and the observed filter. Default: 0.
    ebv : float, array-like, optional
        Selective extinction :math:`E(B-V)` in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with
        :math:`R_V=3.1`. Its shape must be broadcastable to any array-like arguments. Default: 0.
    spectrum_kwargs : dict, optional
        Keyword arguments to be passed to the ``spectrum`` function.
    show : bool, optional
        Plot the chain histories, and display all plots at the end of the script. Default: only save the corner plot
    outpath : str, optional
        Directory to which to save the corner plots. Default: current directory
    nwalkers : int, optional
        Number of walkers (chains) to use for fitting. Default: 10
    burnin_steps : int, optional
        Number of MCMC steps before convergence. This part of the history is discarded. Default: 200
    steps : int, optional
        Number of MCMC steps after convergence. This part of the history is used to calculate parameters. Default: 100.
    save_chains : bool, optional
        If True, save the MCMC chain histories to the directory ``outpath``. Default: only save the corner plot
    use_sigma : bool, optional
        Include an intrinsic scatter parameter. Default: False.
    sigma_type : str, optional
        If 'relative' (default), sigma will be in units of the individual photometric uncertainties.
        If 'absolute', sigma will be in units of the median photometric uncertainty.
    labels : list, optional
        Axis labels for the chain histories and corner plot.
    freq_min, freq_max : float, optional
        Minimum and maximum frequencies for the SED panel in the corner plot. Default: (100., 1000.) or observed range.

    Returns
    -------
    sampler : emcee.EnsembleSampler
        Sampler object containing the results of the MCMC fit
    """
    y = epoch1['lum'].data
    dy = epoch1['dlum'].data
    filtobj = epoch1['filter'].data
    mjdavg = np.mean(epoch1['MJD'].data)

    if spectrum_kwargs is None:
        spectrum_kwargs = {}

    if sigma_type == 'relative':
        sigma_units = dy
    elif sigma_type == 'absolute':
        sigma_units = np.median(dy)
    else:
        raise Exception('sigma_type must either be "relative" or "absolute"')

    def log_posterior(p):
        log_prior = 0.
        for prior, p_i in zip(priors, p):
            log_prior += prior(p_i)
        if np.isinf(log_prior):
            return log_prior
        y_fit = np.array([f.synthesize(spectrum, *p[:-1 if use_sigma else None], z=z, ebv=ebv, **spectrum_kwargs)
                          for f in filtobj])
        sigma = np.sqrt(dy ** 2. + (p[-1] * sigma_units) ** 2.) if use_sigma else dy
        log_likelihood = -0.5 * np.sum(np.log(2 * np.pi * sigma ** 2.) + ((y - y_fit) / sigma) ** 2.)
        return log_prior + log_likelihood

    ndim = len(priors)
    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_posterior)
    pos, _, _ = sampler.run_mcmc(starting_guesses, burnin_steps)

    # Plotting
    if show:
        plot_chain(sampler.chain, labels)
    sampler.reset()
    sampler.run_mcmc(pos, steps)
    if show:
        plot_chain(sampler.chain, labels)

    os.makedirs(outpath, exist_ok=True)
    if save_chains:
        chain_filename = os.path.join(outpath, f'{mjdavg:.3f}.npy')
        np.save(chain_filename, sampler.flatchain)

    f4 = spectrum_corner(spectrum, epoch1, sampler.flatchain, z, ebv, spectrum_kwargs, use_sigma, labels,
                         freq_min=freq_min, freq_max=freq_max, save_plot_as=os.path.join(outpath, f'{mjdavg:.3f}.pdf'))
    if show:
        plt.show()
    else:
        plt.close(f4)

    return sampler


def spectrum_corner(spectrum, epoch1, sampler_flatchain, z=0., ebv=0., spectrum_kwargs=None, use_sigma=False,
                    labels=None, freq_min=100., freq_max=1000., save_plot_as=''):
    """
    Plot the posterior distributions in a corner (pair) plot, with an inset showing the observed and model SEDs.

    Parameters
    ----------
    spectrum : function
        Function describing the spectrum. The first argument must be frequency in THz, and it must return spectral
        luminosity in watts per hertz. For a blackbody, you can use :func:`.models.planck_fast`.
    epoch1 : lightcurve_fitting.lightcurve.LC
        A single "epoch" of photometry that defines the observed spectral energy distribution
    sampler_flatchain : array-like
        2D array containing the aggregated MCMC chain histories
    z : float, optional
        Redshift between the emission source and the observed filter. Default: 0.
    ebv : float, array-like, optional
        Selective extinction :math:`E(B-V)` in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with
        :math:`R_V=3.1`. Its shape must be broadcastable to any array-like arguments. Default: 0.
    spectrum_kwargs : dict, optional
        Keyword arguments to be passed to the ``spectrum`` function.
    use_sigma : bool, optional
        Include an intrinsic scatter parameter. Default: False.
    labels : list, optional
        Axis labels for the chain histories and corner plot.
    freq_min, freq_max : float, optional
        Minimum and maximum frequencies for the SED panel in the corner plot. Default: (100., 1000.) or observed range.
    save_plot_as : str, optional
        Filename to which to save the resulting plot

    Returns
    -------
    f4 : matplotlib.pyplot.Figure
        Figure containing the corner and SED plots.
    """

    ndim = sampler_flatchain.shape[-1]
    if spectrum_kwargs is None:
        spectrum_kwargs = {}

    f4 = corner.corner(sampler_flatchain, labels=labels)

    ps = sampler_flatchain[np.random.choice(sampler_flatchain.shape[0], 100)].T
    xmin = min(freq_min, epoch1['filter'].max().freq_eff.value)
    xmax = max(freq_max, epoch1['filter'].min().freq_eff.value)
    xfit = np.arange(xmin, xmax)
    freq = xfit * (1. + z)
    yfit = spectrum(freq, *ps[:-1 if use_sigma else None], **spectrum_kwargs) * extinction_law(freq, ebv)
    yscale = 10. ** np.floor(np.log10(yfit.max()))

    # add tick labels and axis labels to the top right figure so tight_layout works
    all_axes = np.reshape(f4.get_axes(), (ndim, ndim))
    top_right = all_axes[0, -1]
    top_right.set_frame_on(True)
    top_right.xaxis.set_major_locator(plt.AutoLocator())
    top_right.xaxis.tick_top()
    top_right.set_xlabel('Frequency (THz)')
    top_right.xaxis.set_label_position('top')
    top_right.yaxis.set_major_locator(plt.AutoLocator())
    top_right.yaxis.tick_right()
    top_right.set_ylabel(f'Luminosity $L_\\nu$ (10$^{{{np.log10(yscale):.0f}}}$ W Hz$^{{-1}}$)')
    top_right.yaxis.set_label_position('right')
    f4.tight_layout(h_pad=0.05, w_pad=0.05)

    bottom_left = all_axes[ndim // 2 - 1, (ndim + 1) // 2]
    if bottom_left is top_right:
        ax = top_right
    else:
        # add a new axis that takes up the entire top right of the figure
        bbox0 = bottom_left.bbox.transformed(f4.transFigure.inverted())
        bbox1 = top_right.bbox.transformed(f4.transFigure.inverted())
        ax = f4.add_axes([bbox0.xmin, bbox0.ymin, bbox1.xmax - bbox0.xmin, bbox1.ymax - bbox0.ymin])
        ax.xaxis.tick_top()
        ax.set_xlabel('Frequency (THz)')
        ax.xaxis.set_label_position('top')
        ax.yaxis.tick_right()
        ax.set_ylabel(f'Luminosity $L_\\nu$ (10$^{{{np.log10(yscale):.0f}}}$ W Hz$^{{-1}}$)')
        ax.yaxis.set_label_position('right')

        # turn the original top right axis back off
        top_right.set_frame_on(False)
        top_right.xaxis.set_major_locator(plt.NullLocator())
        top_right.set_xlabel('')
        top_right.yaxis.set_major_locator(plt.NullLocator())
        top_right.set_ylabel('')

    for row in epoch1:
        ax.errorbar(row['freq'], row['lum'] / yscale, row['dlum'] / yscale, marker='o', **row['filter'].plotstyle)
    ax.plot(xfit, yfit.T / yscale, color='k', alpha=0.05)

    if save_plot_as:
        f4.savefig(save_plot_as)
        print('saving figure as ' + save_plot_as)

    return f4


def plot_bolometric_results(t0, save_plot_as=None):
    """
    Plot the parameters and bolometric light curves that result from fitting the spectral energy distribution

    Parameters
    ----------
    t0 : lightcurve_fitting.lightcurve.LC
        Table containing the results from fitting the spectral energy distribution
    save_plot_as : str, optional
        Filename to which to save the plot.

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object containing the plot
    """
    fig, axarr = plt.subplots(3, figsize=(6, 12), sharex=True)

    axarr[0].errorbar(t0['MJD'], t0['lum'], t0['dlum'], marker='.', ls='none', color='k', label='bolometric, curve_fit')
    axarr[0].plot(t0['MJD'], t0['L_opt'], marker='.', ls='none', color='C0', label='pseudobolometric, curve_fit')
    axarr[0].errorbar(t0['MJD'], t0['L_mcmc'], (t0['dL_mcmc0'], t0['dL_mcmc1']), marker='.', ls='none', color='C1',
                      label='pseudobolometric, MCMC')
    axarr[0].plot(t0['MJD'], t0['L_int'], marker='.', ls='none', color='C2', label='pseudobolometric, integration')
    axarr[0].legend()

    axarr[0].set_yscale('log')
    axarr[0].set_ylabel('Luminosity (W)')

    axarr[1].errorbar(t0['MJD'], t0['radius'], t0['dradius'], color='C0', marker='.', ls='none')
    axarr[1].errorbar(t0['MJD'], t0['radius_mcmc'], (t0['dradius0'], t0['dradius1']), color='C1', marker='.', ls='none')
    axarr[1].set_ylabel('Radius ($1000 R_\\odot$)')

    axarr[2].errorbar(t0['MJD'], t0['temp'], t0['dtemp'], color='C0', marker='.', ls='none')
    axarr[2].errorbar(t0['MJD'], t0['temp_mcmc'], (t0['dtemp0'], t0['dtemp1']), color='C1', marker='.', ls='none')
    axarr[2].set_ylabel('Temperature (kK)')
    axarr[2].set_xlabel('MJD')

    fig.tight_layout()
    if save_plot_as is not None:
        fig.savefig(save_plot_as)

    return fig


def group_by_epoch(lc, res=1., also_group_by=()):
    """
    Group a light curve into epochs that will be treated as single spectral energy distributions.

    Parameters
    ----------
    lc : lightcurve_fitting.lightcurve.LC
        Table containing the observed photometry. Must contain times in an "MJD" column.
    res : float, optional
        Approximate resolution for grouping, in days. Default: 1 day.

    Returns
    -------
    epochs.groups : astropy.table.TableGroups
        Iterable containing single-epoch spectral energy distributions
    """
    x = lc['MJD'].data / res
    frac = np.median(x - np.trunc(x))
    lc['bin'] = np.round(x - frac + np.round(frac)) * res
    group_by = ['bin'] + list(also_group_by)
    for col in also_group_by:
        if np.ma.is_masked(lc[col]):
            lc[col] = lc[col].filled()
    epochs = lc.group_by(group_by)
    return epochs.groups


sigma_sb = const.sigma_sb.to(u.W / (1000. * u.Rsun) ** 2 / u.kK ** 4).value


def stefan_boltzmann(temp, radius, dtemp, drad, covTR):
    """
    Calculate blackbody luminosity and associated uncertainty using the Stefan-Boltzmann law

    Parameters
    ----------
    temp : float, array-like
        Temperature in kilokelvins
    radius : float, array-like
        Radius in units of 1000 solar radii
    dtemp : float, array-like
        Uncertainty in the temperature in kilokelvins
    drad : float, array-like
        Uncertainty in the radius in units of 1000 solar radii
    covTR : float, array-like
        Covariance between the temperature and radius

    Returns
    -------
    lum : float, array-like
        Luminosity in watts
    dlum : float, array-like
        Uncertainty in the luminosity in watts
    """
    lum = 4 * np.pi * radius ** 2 * sigma_sb * temp ** 4
    dlum = 8 * np.pi * sigma_sb * (radius ** 2 * temp ** 8 * drad ** 2
                                   + 4 * radius ** 4 * temp ** 6 * dtemp ** 2
                                   + 4 * radius ** 3 * temp ** 7 * covTR) ** 0.5
    return lum, dlum


def median_and_unc(x, perc_contained=68.):
    """
    Calculate the equal-tailed credible interval, centered on the median, for a sample of data

    Parameters
    ----------
    x : array-like
        The data sample
    perc_contained : float, optional
        The percentage of the probability to be contained in the interval. Default: 68% (1σ)

    Returns
    -------
    median : float
        The median of ``x``
    lower : float
        The lower boundary of the credible interval
    upper : float
        The upper boundary of the credible interval
    """
    q = 50. + np.array([-perc_contained / 2., 0., perc_contained / 2.])
    percentiles = np.percentile(x, q, axis=0)
    median = percentiles[1]
    lower, upper = np.diff(percentiles, axis=0)
    return median, lower, upper


def blackbody_lstsq(epoch1, z, p0=None, T_range=(1., 100.), R_range=(0.01, 1000.), cutoff_freq=np.inf):
    """
    Fit a blackbody spectrum to a spectral energy distribution using :math:`χ^2` minimization

    Parameters
    ----------
    epoch1 : lightcurve_fitting.lightcurve.LC
        A single "epoch" of photometry that defines the observed spectral energy distribution
    z : float
        Redshift between the blackbody (rest frame) and the filters (observed frame)
    p0 : list, tuple, array-like, optional
        Initial guess for [temperature (kK), radius (1000 Rsun)]. Default: ``[10., 10.]``
    T_range : tuple, list, array-like, optional
        Range of allowed temperatures (in kilokelvins) in the prior. Default: ``(1., 100.)``
    R_range : tuple, list, array-like, optional
        Range of allowed radii (in 1000 solar radii) in the prior. Default: ``(0.01, 1000.)``
    cutoff_freq : float, optional
        Cutoff frequency (in terahertz) for a modified blackbody spectrum (see https://doi.org/10.3847/1538-4357/aa9334)

    Returns
    -------
    temp : float
        Best-fit blackbody temperature in kilokelvins
    radius : float
        Best-fit blackbody radius in units of 1000 solar radii
    dtemp : float
        Uncertainty in the temperature in kilokelvins
    drad : float
        Uncertainty in the radius in units of 1000 solar radii
    lum : float
        Blackbody luminosity implied by the best-fit temperature and radius
    dlum : float
        Uncertainty in the blackbody luminosity
    L_opt : float
        Pseudobolometric luminosity from integrating the best-fit blackbody spectrum over the U-I filters
    """
    if p0 is None:
        p0 = [10., 10.]

    def planck_cutoff(nu, T, R):
        return planck_fast(nu, T, R, cutoff_freq)

    with warnings.catch_warnings():
        if len(epoch1) <= 2:
            warnings.simplefilter('ignore', OptimizeWarning)
        p0, cov = curve_fit(planck_cutoff, epoch1['freq'] * (1. + z), epoch1['lum'], p0=p0,
                            bounds=([T_range[0], R_range[0]], [T_range[1], R_range[1]]))
    temp, radius = p0
    dtemp, drad = np.sqrt(np.diag(cov))
    lum, dlum = stefan_boltzmann(temp, radius, dtemp, drad, cov[0, 1])
    L_opt = pseudo(temp, radius, z, cutoff_freq=cutoff_freq)
    return temp, radius, dtemp, drad, lum, dlum, L_opt


def integrate_sed(epoch1):
    """
    Directly integrate a spectral energy distribution using the trapezoidal rule

    Parameters
    ----------
    epoch1 : lightcurve_fitting.lightcurve.LC
        A single "epoch" of photometry that defines the observed spectral energy distribution

    Returns
    -------
    L_int : float
        Luminosity in watts
    """
    epoch1.sort('freq')
    freqs = np.insert(epoch1['freq'], 0, epoch1['freq'][0] - epoch1['dfreq'][0])
    lums = np.insert(epoch1['lum'], 0, 0)
    freqs = np.append(freqs, epoch1['freq'][-1] + epoch1['dfreq'][-1])
    lums = np.append(lums, 0)
    L_int = np.trapz(lums * epoch1['lum'].unit, freqs * epoch1['freq'].unit).to(u.W).value
    return L_int


def calc_colors(epoch1, colors):
    """
    Calculate colors from a spectral energy distribution

    Parameters
    ----------
    epoch1 : lightcurve_fitting.lightcurve.LC
        A single "epoch" of photometry that defines the observed spectral energy distribution
    colors : list
        A list of colors to calculate, e.g., ``['U-B', 'B-V', 'g-r', 'r-i']``

    Returns
    -------
    mags : list
        A list of the calculated colors
    dmags : list
        A list of uncertainties on the calculated colors, from propagating errors on the photometry points
    lolims : list
        A list of booleans: True if the first filter is a nondetection or is absent from ``epoch1``
    uplims : list
        A list of booleans: True if the second filter is a nondetection or is absent from ``epoch1``
    """
    mags = []
    dmags = []
    lolims = []
    uplims = []
    for color in colors:
        f0, f1 = [filtdict[f] for f in color.split('-')]
        if f0 in epoch1['filter'] and f1 in epoch1['filter']:
            m0, dm0, n0 = epoch1.where(filter=f0)[['absmag', 'dmag', 'nondet']][0]
            m1, dm1, n1 = epoch1.where(filter=f1)[['absmag', 'dmag', 'nondet']][0]
            if n0 and n1:
                m0_m1 = np.nan
            else:
                m0_m1 = m0 - m1
            dm0_m1 = (dm0 ** 2. + dm1 ** 2.) ** 0.5
            mags.append(m0_m1)
            dmags.append(dm0_m1)
            lolims.append(n0)
            uplims.append(n1)
        else:
            mags.append(np.nan)
            dmags.append(np.nan)
            lolims.append(True)
            uplims.append(True)
    return mags, dmags, lolims, uplims


def plot_color_curves(t, colors=None, fmt='o', limit_length=0.1, xcol='MJD'):
    """
    Plot the color curves calculated by :func:`.calculate_bolometric`.

    Parameters
    ----------
    t : lightcurve_fitting.lightcurve.LC
        Output table from :func:`.calculate_bolometric`
    colors : list, optional
        List of colors to plot, e.g., ``['g-r', 'r-i']``. By default, plot all recognizable colors.
    fmt : str, optional
        Format string, passed to :func:`matplotlib.pyplot.errorbar`
    limit_length : float, optional
        Length (in data units) of the upper and lower limit markers

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object containing the plot
    """
    if colors is None:
        colors = []
        for col in t.colnames:
            if col.split('-')[0] in filtdict and not (t.has_masked_values and t.mask[col].all()):
                colors.append(col)
    fig = plt.figure()
    for c in colors:
        dcolor_colname = f'd({c})'
        if t.has_masked_values and t.mask[dcolor_colname].any():
            dcolor = t[dcolor_colname].filled(limit_length)
        else:
            dcolor = t[dcolor_colname]
        plt.errorbar(t[xcol], t[c], dcolor, (t[f'd{xcol}0'], t[f'd{xcol}1']), fmt=fmt,
                     lolims=t[f'lolims({c})'], uplims=t[f'uplims({c})'], label=f'${c}$')
    plt.xlabel(xcol)
    plt.ylabel('Color (mag)')
    plt.legend()
    return fig


def calculate_bolometric(lc, z=0., outpath='.', res=1., nwalkers=10, burnin_steps=200, steps=100, priors=None,
                         save_table_as=None, min_nfilt=3, cutoff_freq=np.inf, show=False, colors=None, do_mcmc=True,
                         save_chains=False, use_sigma=False, sigma_type='relative', also_group_by=()):
    """
    Calculate the full bolometric light curve from a table of broadband photometry

    Parameters
    ----------
    lc : lightcurve_fitting.lightcurve.LC
        Table of broadband photometry including columns "MJD", "mag", "dmag", "filt"
    z : float, optional
        Redshift between the emission source and the observed filter. Default: 0.
    outpath : str, optional
        Directory to which to save the corner plots and MCMC chains. Default: current directory
    res : float, optional
        Approximate resolution for grouping, in days. Default: 1 day.
    nwalkers : int, optional
        Number of walkers (chains) to use for fitting. Default: 10
    burnin_steps : int, optional
        Number of MCMC steps before convergence. This part of the history is discarded. Default: 200
    steps : int, optional
        Number of MCMC steps after convergence. This part of the history is used to calculate paramers. Default: 100.
    priors : list, optional
        Prior probability distributions for temperature (in kilokelvins) and radius (in 1000 solar radii). Available
        priors: :class:`.models.UniformPrior`, :class:`.models.LogUniformPrior`, :class:`.models.GaussianPrior`.
        Default: ``T = Uniform(1., 100.)``, ``R = LogUniform(0.01, 1000.)``, ``σ = Gaussian(0., 10., stddev=1.)``.
    save_table_as : str, optional
        Filename to which to save the output table of blackbody parameters and bolometric luminosities
    min_nfilt : int, optional
        Minimum number of distinct observed filters required to calculate a luminosity. Default: 3
    cutoff_freq : float, optional
        Cutoff frequency (in terahertz) for a modified blackbody spectrum (see https://doi.org/10.3847/1538-4357/aa9334)
    show : bool, optional
        Plot the chain histories, and display all plots at the end of the script. Default: only save the corner plot
    colors : list, optional
        A list of colors to calculate, e.g., ``['U-B', 'B-V', 'g-r', 'r-i']``
    do_mcmc : bool, optional
        If True (default), also fit the spectral energy distribution with an MCMC routine. This is slower but gives
        more realistic uncertainties.
    save_chains : bool, optional
        If True, save the MCMC chain histories to the directory ``outpath``. Default: only save the corner plot
    use_sigma : bool, optional
        Include an intrinsic scatter parameter in the MCMC fit. Default: False.
    sigma_type : str, optional
        If 'relative' (default), sigma will be in units of the individual photometric uncertainties.
        If 'absolute', sigma will be in units of the median photometric uncertainty.
    also_group_by : list, optional
        Group by these columns in addition to epoch

    Returns
    -------
    t0 : lightcurve_fitting.lightcurve.LC
        Table containing the blackbody parameters, bolometric luminosities, and (optionally) colors
    """

    if colors is None:
        colors = []

    use_src = 'source' in lc.colnames
    t0 = LC(names=['MJD', 'dMJD0', 'dMJD1',
                   'temp', 'radius', 'dtemp', 'dradius',  # best fit from scipy.curve_fit
                   'lum', 'dlum',  # total bolometric luminosity from scipy.curve_fit
                   'L_opt',  # pseudobolometric luminosity from scipy.curve_fit
                   'temp_mcmc', 'radius_mcmc', 'dtemp0', 'dtemp1', 'dradius0', 'dradius1',  # best fit from MCMC
                   'L_mcmc', 'dL_mcmc0', 'dL_mcmc1',  # pseudobolometric luminosity from MCMC
                   'L_int',  # pseudobolometric luminosity from direct integration of the SED
                   'npoints']
            + colors + ['d({})'.format(c) for c in colors] + ['lolims({})'.format(c) for c in colors]
            + ['uplims({})'.format(c) for c in colors] + ['filts'] + (['source'] if use_src else []),
            dtype=[float, float, float, float, float, float, float, float, float, float, float, float, float, float,
                   float, float, float, float, float, float, int]
            + [float] * 2 * len(colors) + [bool] * 2 * len(colors) + ['S6'] + ([lc['source'].dtype] if use_src else []),
            masked=True)

    if priors is None:
        priors = [UniformPrior(1., 100.), LogUniformPrior(0.01, 1000.)]
        if use_sigma:
            priors.append(GaussianPrior(0., 10.))

    sampler = None
    lc = lc[np.isfinite(lc['dmag']) & (lc['dmag'] > 0.)]
    for epoch1 in group_by_epoch(lc, res, also_group_by):
        epoch1.sn = lc.sn
        epoch1.calcFlux()
        epoch1 = epoch1.bin(delta=np.inf)
        epoch1.calcMag()
        epoch1.calcAbsMag()
        epoch1.calcLum()

        epoch1['freq'] = u.Quantity([f.freq_eff for f in epoch1['filter']])
        epoch1['dfreq'] = u.Quantity([f.dfreq for f in epoch1['filter']])

        epoch1['lum'].unit = u.W / u.Hz
        epoch1['dlum'].unit = u.W / u.Hz

        filts = set(epoch1.where(nondet=False)['filter'].data)
        nfilt = len(filts)
        if nfilt < min_nfilt:
            continue

        if nfilt > 1:
            p0 = np.array([10., 10.])
        elif sampler is not None:
            priors[0] = gaussian_kde(sampler.flatchain[:, 0]).logpdf
            p0 = np.median(sampler.flatchain, axis=0)
        else:
            continue

        mjdavg, dmjd0, dmjd1 = median_and_unc(epoch1['MJD'], 100.)
        filtstr = ''.join([f.char for f in sorted(filts)])

        # blackbody - least squares
        T_range = (priors[0].p_min, priors[0].p_max)
        R_range = (priors[1].p_min, priors[1].p_max)
        try:
            temp, radius, dtemp, drad, lum, dlum, L_opt = blackbody_lstsq(epoch1, z, p0, T_range, R_range, cutoff_freq)
            p0 = np.array([temp, radius])
        except RuntimeError:  # optimization failed
            temp = radius = dtemp = drad = lum = dlum = L_opt = np.nan

        rng = np.random.default_rng()
        starting_guesses = rng.normal(size=(nwalkers, 2)) + p0
        starting_guesses[starting_guesses <= 0.] = 1.
        labels = ['T (kK)', 'R (1000 R$_\\odot$)']
        if use_sigma:
            starting_guesses = np.append(starting_guesses, np.abs(rng.normal(size=(nwalkers, 1))), axis=1)
            labels.append('$\\sigma$')

        # blackbody - MCMC
        try:
            if not do_mcmc:
                raise ValueError
            spectrum_kwargs = {'cutoff_freq': cutoff_freq}
            sampler = spectrum_mcmc(planck_fast, epoch1, priors, starting_guesses, z=z, spectrum_kwargs=spectrum_kwargs,
                                    outpath=outpath, nwalkers=nwalkers, burnin_steps=burnin_steps, steps=steps,
                                    show=show, save_chains=save_chains, use_sigma=use_sigma, sigma_type=sigma_type,
                                    labels=labels)
            L_mcmc_opt = pseudo(sampler.flatchain[:, 0], sampler.flatchain[:, 1], z, cutoff_freq=cutoff_freq)
            (T_mcmc, R_mcmc), (dT0_mcmc, dR0_mcmc), (dT1_mcmc, dR1_mcmc) = median_and_unc(sampler.flatchain[:, :2])
            L_mcmc, dL_mcmc0, dL_mcmc1 = median_and_unc(L_mcmc_opt)
        except ValueError as e:
            print(e)
            T_mcmc = R_mcmc = dT0_mcmc = dR0_mcmc = dT1_mcmc = dR1_mcmc = L_mcmc = dL_mcmc0 = dL_mcmc1 = np.nan

        # direct integration
        L_int = integrate_sed(epoch1)

        # color calculation
        if colors is None:
            colors = []
        color_mags, color_dmags, color_lolims, color_uplims = calc_colors(epoch1, colors)

        row = [mjdavg, dmjd0, dmjd1,
               temp, radius, dtemp, drad, lum, dlum, L_opt,
               T_mcmc, R_mcmc, dT0_mcmc, dT1_mcmc, dR0_mcmc, dR1_mcmc, L_mcmc, dL_mcmc0, dL_mcmc1,
               L_int, nfilt] + color_mags + color_dmags
        row_bool = color_lolims + color_uplims
        row_string = [filtstr] + ([epoch1['source'][0]] if use_src else [])
        mask = np.concatenate([np.isnan(row), np.zeros_like(row_bool, dtype=bool),
                               ~np.array([bool(rs) for rs in row_string])])
        t0.add_row(row + row_bool + row_string, mask=mask)

    if save_table_as is not None and t0:
        t0.write(save_table_as, format='ascii.fixed_width_two_line', overwrite=True)

    return t0



import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u
import emcee
import corner
from .models import CompanionShocking, CompanionShocking2, scale_sifto, UniformPrior
from pkg_resources import resource_filename
import warnings

PRIOR_WARNING = 'The p_max/p_min keywords are deprecated. Use the priors keyword instead.'


def lightcurve_mcmc(lc, model, priors=None, p_min=None, p_max=None, p_lo=None, p_up=None,
                    nwalkers=100, nsteps=1000, nsteps_burnin=1000, model_kwargs=None,
                    show=False, save_plot_as='', save_sampler_as='', use_sigma=False, sigma_type='relative'):
    """
    Fit an analytical model to observed photometry using a Markov-chain Monte Carlo routine

    Parameters
    ----------
    lc : lightcurve_fitting.lightcurve.LC
        Table of broadband photometry including columns "MJD", "mag", "dmag", "filt"
    model : lightcurve_fitting.models.Model
        The model to fit to the light curve. Available models: :class:`.models.ShockCooling`,
        :class:`.models.ShockCooling2`, :class:`.models.ShockCooling3`, :class:`.models.CompanionShocking`,
        :class:`.models.CompanionShocking2`
    priors : list, optional
        Prior probability distributions for each model parameter. Available priors:
        :class:`.models.UniformPrior` (default), :class:`.models.LogUniformPrior`, :class:`.models.GaussianPrior`
    p_min, p_max : list, optional
        DEPRECATED: Use `priors` instead
    p_lo : list
        Lower bounds on the starting guesses for each paramter
    p_up : list
        Upper bounds on the starting guesses for each parameter
    nwalkers : int, optional
        Number of walkers (chains) for the MCMC routine. Default: 100
    nsteps : int, optional
        Number of steps (iterations) for the MCMC routine, excluding burn-in. Default: 1000
    nsteps_burnin : int, optional
        Number of steps (iterations) for the MCMC routine during burn-in. Default: 1000
    model_kwargs : dict, optional
        Keyword arguments to be passed to the model
    show : bool, optional
        If True, plot and display the chain histories
    save_plot_as : str, optional
        Save a plot of the chain histories to this filename
    save_sampler_as : str, optional
        Save the aggregated chain histories to this filename
    use_sigma : bool, optional
        If True, treat the last parameter as an intrinsic scatter parameter that does not get passed to the model
    sigma_type : str, optional
        If 'relative' (default), sigma will be in units of the individual photometric uncertainties.
        If 'absolute', sigma will be in units of the median photometric uncertainty.

    Returns
    -------
    sampler : emcee.EnsembleSampler
        EnsembleSampler object containing the results of the fit
    """

    if model_kwargs is None:
        model_kwargs = {}

    if model.output_quantity == 'flux':
        lc.calcFlux()
    elif model.output_quantity == 'lum':
        lc.calcAbsMag()
        lc.calcLum()

    f = lc['filter'].data
    t = lc['MJD'].data
    y = lc[model.output_quantity].data
    dy = lc['d'+model.output_quantity].data

    if model in [CompanionShocking, CompanionShocking2]:
        scale_sifto(lc)

    if use_sigma:
        model.axis_labels.append('$\\sigma$')

    ndim = model.nparams + use_sigma

    # DEPRECATED
    if p_min is None:
        p_min = np.tile(-np.inf, ndim)
    elif len(p_min) == ndim:
        p_min = np.array(p_min, float)
        warnings.warn(PRIOR_WARNING)
    else:
        raise Exception(PRIOR_WARNING)

    # DEPRECATED
    if p_max is None:
        p_max = np.tile(np.inf, ndim)
    elif len(p_max) == ndim:
        p_max = np.array(p_max, float)
        warnings.warn(PRIOR_WARNING)
    else:
        raise Exception(PRIOR_WARNING)

    if p_lo is None:
        p_lo = p_min
    elif len(p_lo) == ndim:
        p_lo = np.array(p_lo, float)
    else:
        raise Exception('p_lo must have length {:d}'.format(ndim))

    if len(p_up) == ndim:
        p_up = np.array(p_up, float)
    else:
        raise Exception('p_up must have length {:d}'.format(ndim))

    if priors is None:
        priors = [UniformPrior(p0, p1) for p0, p1 in zip(p_min, p_max)]
    elif len(priors) != ndim:
        raise Exception('priors must have length {:d}'.format(ndim))

    for param, prior, p0, p1 in zip(model.input_names, priors, p_lo, p_up):
        if p0 < prior.p_min:
            raise Exception(f'starting guess for {param} (p_lo = {p0}) is outside prior (p_min = {prior.p_min})')
        if p1 > prior.p_max:
            raise Exception(f'starting guess for {param} (p_up = {p1}) is outside prior (p_max = {prior.p_max})')

    if sigma_type == 'relative':
        sigma_units = dy
    elif sigma_type == 'absolute':
        sigma_units = np.median(dy)
    else:
        raise Exception('sigma_type must either be "relative" or "absolute"')

    def log_posterior(p):
        log_prior = 0.
        for prior, p_i in zip(priors, p):
            log_prior += prior(p_i)
        if np.isinf(log_prior):
            return log_prior
        y_fit = model(t, f, *p, **model_kwargs)
        sigma = np.sqrt(dy ** 2. + (p[-1] * sigma_units) ** 2.) if use_sigma else dy
        log_likelihood = -0.5 * np.sum(np.log(2 * np.pi * sigma ** 2.) + ((y - y_fit) / sigma) ** 2.)
        return log_prior + log_likelihood

    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_posterior)

    starting_guesses = np.random.rand(nwalkers, ndim) * (p_up - p_lo) + p_lo
    pos, _, _ = sampler.run_mcmc(starting_guesses, nsteps_burnin, progress=True, progress_kwargs={'desc': ' Burn-in'})

    if show or save_plot_as:
        fig, ax = plt.subplots(ndim, 2, figsize=(12., 2. * ndim))
        ax1 = ax[:, 0]
        for i in range(ndim):
            ax1[i].plot(sampler.chain[:, :, i].T, 'k', alpha=0.2)
            ax1[i].set_ylabel(model.axis_labels[i])
        ax1[0].set_title('During Burn In')
        ax1[-1].set_xlabel('Step Number')

    sampler.reset()
    sampler.run_mcmc(pos, nsteps, progress=True, progress_kwargs={'desc': 'Sampling'}, skip_initial_state_check=True)
    if save_sampler_as:
        np.save(save_sampler_as, sampler.flatchain)
        print('saving sampler.flatchain as ' + save_sampler_as)

    if show or save_plot_as:
        ax2 = ax[:, 1]
        for i in range(ndim):
            ax2[i].plot(sampler.chain[:, :, i].T, 'k', alpha=0.2)
            ax2[i].set_ylabel(model.axis_labels[i])
            ax2[i].yaxis.set_label_position('right')
            ax2[i].yaxis.tick_right()
        ax2[0].set_title('After Burn In')
        ax2[-1].set_xlabel('Step Number')
        fig.tight_layout()

        if save_plot_as:
            print('saving chain plot as ' + save_plot_as)
            fig.savefig(save_plot_as)

        if show:
            plt.show()

    return sampler


def lightcurve_corner(lc, model, sampler_flatchain, model_kwargs=None,
                      num_models_to_plot=100, lcaxis_posn=(0.7, 0.55, 0.2, 0.4),
                      filter_spacing=0.5, tmin=None, tmax=None, t0_offset=None, save_plot_as='', ycol=None,
                      textsize='medium', param_textsize='large'):
    """
    Plot the posterior distributions in a corner (pair) plot, with an inset showing the observed and model light curves.

    Parameters
    ----------
    lc : lightcurve_fitting.lightcurve.LC
        Table of broadband photometry including columns "MJD", "mag", "dmag", "filt"
    model : lightcurve_fitting.models.Model
        The model that was fit to the light curve.
    sampler_flatchain : array-like
        2D array containing the aggregated MCMC chain histories
    model_kwargs : dict, optional
        Keyword arguments to be passed to the model
    num_models_to_plot : int, optional
        Number of model realizations to plot in the light curve inset. Default: 100
    lcaxis_posn : tuple, optional
        Light curve inset position and size specification in figure units: (left, bottom, width, height)
    filter_spacing : float, optional
        Spacing between filters in the light curve inset, in units determined by the order of magnitude of the
        luminosities. Default: 0.5
    tmin, tmax : float, optional
        Starting and ending times for which to plot the models in the light curve inset. Default: determined by the
        time range of the observed light curve.
    t0_offset : float, optional
        Reference time for the explosion time in the corner plot. Default: the earliest explosion time in
        `sampler_flatchain`, rounded down.
    save_plot_as : str, optional
        Filename to which to save the resulting plot
    ycol : str, optional
        Quantity to plot on the light curve inset. Choices: "lum", "flux", or "absmag". Default: model.output_quantity
    textsize : str, optional
        Font size for the x- and y-axis labels, as well as the tick labels. Default: 'medium'
    param_textsize : str, optional
        Font size for the parameter text. Default: 'large'

    Returns
    -------
    fig : matplotlib.pyplot.Figure
        Figure object containing the plot
    corner_ax : array-like
        Array of matplotlib.pyplot.Axes objects corresponding to the corner plot
    ax : matplotlib.pyplot.Axes
        Axes object for the light curve inset
    """
    if model_kwargs is None:
        model_kwargs = {}
    if ycol is None:
        ycol = model.output_quantity
    plt.style.use(resource_filename('lightcurve_fitting', 'serif.mplstyle'))

    sampler_flatchain_corner = sampler_flatchain.copy()
    if 't_0' in model.input_names:
        i_t0 = model.input_names.index('t_0')
        if t0_offset is None:
            t0_offset = np.floor(sampler_flatchain_corner[:, i_t0].min())
        if t0_offset != 0.:
            sampler_flatchain_corner[:, i_t0] -= t0_offset
            model.axis_labels[i_t0] = '$t_0 - {:.0f}$ (d)'.format(t0_offset)

    fig = corner.corner(sampler_flatchain_corner, labels=model.axis_labels, label_kwargs={'size': textsize})
    corner_axes = np.array(fig.get_axes()).reshape(sampler_flatchain.shape[-1], sampler_flatchain.shape[-1])
    for i in range(sampler_flatchain.shape[-1]):
        corner_axes[i, 0].tick_params(labelsize=textsize)
        corner_axes[-1, i].tick_params(labelsize=textsize)

    for ax in np.diag(corner_axes):
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('none')

    ax = fig.add_axes(lcaxis_posn)
    lightcurve_model_plot(lc, model, sampler_flatchain, model_kwargs, num_models_to_plot, filter_spacing,
                          tmin, tmax, ycol, textsize, ax)

    paramtexts = format_credible_interval(sampler_flatchain, varnames=model.input_names, units=model.units)
    fig.text(0.45, 0.95, '\n'.join(paramtexts), va='top', ha='center', fontdict={'size': param_textsize})
    if save_plot_as:
        fig.savefig(save_plot_as)
        print('saving figure as ' + save_plot_as)

    return fig, corner_axes, ax


def lightcurve_model_plot(lc, model, sampler_flatchain, model_kwargs=None, num_models_to_plot=100, filter_spacing=0.5,
                          tmin=None, tmax=None, ycol=None, textsize='medium', ax=None, mjd_offset=None):
    """
    Plot the observed and model light curves.

    Parameters
    ----------
    lc : lightcurve_fitting.lightcurve.LC
        Table of broadband photometry including columns "MJD", "mag", "dmag", "filt"
    model : lightcurve_fitting.models.Model
        The model that was fit to the light curve.
    sampler_flatchain : array-like
        2D array containing the aggregated MCMC chain histories
    model_kwargs : dict, optional
        Keyword arguments to be passed to the model
    num_models_to_plot : int, optional
        Number of model realizations to plot in the light curve inset. Default: 100
    filter_spacing : float, optional
        Spacing between filters in the light curve inset, in units determined by the order of magnitude of the
        luminosities. Default: 0.5
    tmin, tmax : float, optional
        Starting and ending times for which to plot the models in the light curve inset. Default: determined by the
        time range of the observed light curve.
    ycol : str, optional
        Quantity to plot on the light curve inset. Choices: "lum", "flux", or "absmag". Default: model.output_quantity
    textsize : str, optional
        Font size for the x- and y-axis labels, as well as the tick labels. Default: 'medium'
    ax : matplotlib.pyplot.Axes
        Axis on which to plot the light curves
    mjd_offset : float, optional
        Reference time on the horizontal axis of the light curve inset. Default: determined by the starting time of
        the model light curve.
    """
    if ax is None:
        ax = plt.axes()

    choices = np.random.choice(sampler_flatchain.shape[0], num_models_to_plot)
    ps = sampler_flatchain[choices].T

    if tmin is None:
        tmin = np.min(lc['MJD'])
    if tmax is None:
        tmax = np.max(lc['MJD'])
    xfit = np.arange(tmin, tmax, 0.1)
    ufilts = np.unique(lc['filter'])
    y_fit = model(xfit, ufilts, *ps, **model_kwargs)

    if mjd_offset is None:
        mjd_offset = np.floor(tmin)
    if ycol == 'lum':
        dycol = 'dlum'
        yscale = 10. ** np.round(np.log10(y_fit.max()))
        ylabel = 'Luminosity $L_\\nu$ (10$^{{{:.0f}}}$ erg s$^{{-1}}$ Hz$^{{-1}}$) + Offset'.format(
            np.log10(yscale) + 7)  # W --> erg / s
    elif ycol == 'absmag':
        dycol = 'dmag'
        yscale = 1.
        ylabel = 'Absolute Magnitude + Offset'
        y_fit = [[[filt.M0]] for filt in ufilts] - 2.5 * np.log10(y_fit)
        ax.invert_yaxis()
    elif ycol == 'flux':
        dycol = 'dflux'
        yscale = 10. ** np.round(np.log10(y_fit.max()))
        ylabel = 'Flux $F_\\nu$ (10$^{{{:.0f}}}$ erg s$^{{-1}}$ m$^{{-2}}$ Hz$^{{-1}}$) + Offset'.format(
            np.log10(yscale) + 7)  # W --> erg / s
    else:
        raise ValueError(f'ycol="{ycol}" is not recognized. Use "lum", "absmag", "flux".')
    offset = -len(ufilts) // 2 * filter_spacing
    for filt, yfit in zip(ufilts, y_fit):
        offset += filter_spacing
        lc_filt = lc.where(filter=filt)
        ax.errorbar(lc_filt['MJD'] - mjd_offset, lc_filt[ycol] / yscale + offset, lc_filt[dycol] / yscale,
                    ls='none', marker='o', **filt.plotstyle)
        ax.plot(xfit - mjd_offset, yfit / yscale + offset, color=filt.linecolor, alpha=0.05)
        txt = f'${filt.name}{offset:+g}$' if filt.italics else rf'$\mathrm{{{filt.name}}}{offset:+g}$'
        ax.text(1.03, yfit[-1, 0] / yscale + offset, txt, color=filt.textcolor, fontdict={'size': textsize},
                ha='left', va='center', transform=ax.get_yaxis_transform())
    ax.set_xlabel('MJD $-$ {:f}'.format(mjd_offset).rstrip('0').rstrip('.'), size=textsize)
    ax.set_ylabel(ylabel, size=textsize)
    ax.tick_params(labelsize=textsize)


def format_credible_interval(x, sigfigs=1, percentiles=(15.87, 50., 84.14), axis=0, varnames=None, units=None):
    """
    Use LaTeX to format an equal-tailed credible interval with a given number of significant figures in the uncertainty

    Parameters
    ----------
    x : array-like
        Data from which to calculate the credible interval
    sigfigs : int, optional
        Number of significant figures in the uncertainty. Default: 1
    percentiles : tuple, optional
        Percentiles for the (lower, center, upper) of the credible interval. Default: ``(15.87, 50., 84.14)``
        (median +/- 1σ)
    axis : int, optional
        Axis of ``x`` along which to calculate the credible intervals. Default: 0
    varnames : list, optional
        Variable names to be equated with the credible intervals. Default: no variable names
    units : list, optional
        Units to be applied to the credible intervals. Default: no units

    Returns
    -------
    paramtexts : str
        The formatted credible intervals
    """
    quantiles = np.percentile(x, percentiles, axis=axis).T
    uncertainties = np.diff(quantiles)
    smaller_unc = np.amin(uncertainties, axis=-1)
    log_unc = np.log10(smaller_unc)
    roundto = sigfigs - np.floor(log_unc).astype(int) - 1

    # catch numbers that will have more sigfigs after rounding
    smaller_unc_round = [np.round(unc, dec) for unc, dec in zip(smaller_unc, roundto)]
    log_unc_round = np.log10(smaller_unc_round)
    roundto = sigfigs - np.floor(log_unc_round).astype(int) - 1

    quantiles = np.atleast_2d(quantiles)
    uncertainties = np.atleast_2d(uncertainties)
    roundto = np.atleast_1d(roundto)
    texstrings = []
    for quant, unc, dec in zip(quantiles, uncertainties, roundto):
        center = np.round(quant[1], dec)
        lower, upper = np.round(unc, dec)
        if dec < 0:
            dec = 0
        if upper == lower:
            texstring = '{{:.{0:d}f}} \\pm {{:.{0:d}f}}'.format(dec).format(center, upper)
        else:
            texstring = '{{:.{0:d}f}}^{{{{+{{:.{0:d}f}}}}}}_{{{{-{{:.{0:d}f}}}}}}'.format(dec).format(center, upper,
                                                                                                      lower)
        texstrings.append(texstring)

    if varnames is None or units is None:
        paramtexts = texstrings
    else:
        paramtexts = []
        for var, value, unit in zip(varnames, texstrings, units):
            if isinstance(unit, u.quantity.Quantity):
                value = '({}) \\times 10^{{{:.1f}}}'.format(value, np.log10(unit.value)).replace('.0}', '}')
                unit = unit.unit
            paramtexts.append('${} = {}$ {:latex_inline}'.format(var, value, unit))

    return paramtexts



#!/usr/bin/env python

import numpy as np
from .lightcurve import LC
from .filters import filtdict
from astropy import constants as const, units as u
from astropy.io import fits, ascii
from astropy.wcs import WCS
from astropy.time import Time
from astropy.table import Table
import argparse
import matplotlib.pyplot as plt
import os
import warnings
import json
import re
import shutil


def removebadcards(hdr):
    """
    Remove problematic entries from a :class:`astropy.io.fits.header.Header`
    """
    for card in hdr.cards:
        try:
            card.verify('fix')
        except fits.verify.VerifyError as e:
            print(e)
            hdr.remove(card.keyword)
        except Exception as e:
            print(e)
            hdr.remove(card.keyword)
    return hdr


def remove_duplicate_wcs(hdr, keep_number=0):
    """
    Remove duplicate world coordinate system header keywords from a :class:`astropy.io.fits.header.Header`
    """
    for key in ['CTYPE1', 'CTYPE2', 'CRPIX1', 'CRPIX2', 'CRVAL1', 'CRVAL2', 'CD1_1', 'CD2_2', 'CD1_2', 'CD2_1']:
        if key in hdr and hdr.count(key) > 1:
            card = hdr.cards[(key, keep_number)]
            hdr.remove(card.keyword, remove_all=True)
            hdr[card.keyword] = (card.value, card.comment)


def readfitsspec(filename, header=False, ext=None):
    """
    Read a spectrum from a FITS file

    Parameters
    ----------
    filename : str
        Filename from which to read the spectrum
    header : bool, optional
        If True, also return the FITS header
    ext : int, str, optional
        FITS extension number or name in which the spectrum is stored

    Returns
    -------
    wl : array-like
        Wavelengths, typically in ångströms
    flux : array-like
        Observed fluxes in erg / (s cm2 angstrom), if units are identifiable
    hdr : astropy.io.fits.header.Header, optional
        FITS header, returned if ``header=True``
    """
    hdulist = fits.open(filename)
    if ext is None:
        for hdu in hdulist:  # try to find SCI extension
            if hdu.header.get('extname') == 'SCI':
                break
        else:  # if no SCI extension, pick the first extension with data
            for hdu in hdulist:
                if hdu.data is not None:
                    break
            else:
                raise Exception('no extensions have any data')
    else:
        hdu = hdulist[ext]
    data = hdu.data
    hdr = hdu.header
    if isinstance(hdu, fits.BinTableHDU):
        wl = data['wavelength']
        flux = data['flux']
    else:
        flux = data.flatten()[:max(data.shape)]
        remove_duplicate_wcs(hdr)  # some problem with Gemini pipeline
        if hdr.get('CUNIT1') in ['Angstroms', 'angstroms', 'deg', 'pixel']:
            hdr['CUNIT1'] = 'Angstrom'  # WCS object needs recognizable units
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            wcs = WCS(removebadcards(hdr), naxis=1, relax=False, fix=False)
        if 'CUNIT1' in hdr:
            wl = wcs.pixel_to_world(np.arange(len(flux))).to(hdr['CUNIT1']).value
        else:
            wl = wcs.wcs_pix2world(np.arange(len(flux)), 0)[0]
    if header:
        return wl, flux, hdr
    else:
        return wl, flux


def convert_spectrum_units(wl, flux, hdr, default_bunit='erg / (Angstrom cm2 s)', default_cunit='Angstrom'):
    """
    Convert a spectrum to standard units, if information is available in the header to establish units (BUNIT & CUNIT1)

    Parameters
    ----------
    wl: array-like
        Input wavelengths
    flux: array-like
        Input fluxes
    hdr: dict-like
        Metadata from which to determine the units of wl and flux
    default_bunit: str, optional
        Units of the output flux. Default: 'erg / (Angstrom cm2 s)'
    default_cunit: str, optional
        Units of the output wavelengths. Default: 'Angstrom'

    Returns
    -------
    wl: numpy.ndarray
        Input wavelengths converted to `default_cunit`
    flux: numpy.ndarray
        Input fluxes converted to `default_bunit`
    """
    bunit = hdr.get('BUNIT', default_bunit)
    if bunit == 'adu':
        bunit = default_bunit
    if 'Angstrom' not in bunit:
        bunit = bunit.replace('Ang', 'Angstrom')
    if 'Angstrom' not in bunit:
        bunit = bunit.replace('A', 'Angstrom')
    cunit = hdr.get('CUNIT1', hdr.get('XUNITS', default_cunit))
    if cunit.lower() == 'angstroms':
        cunit = cunit.rstrip('s')
    wl = u.Quantity(wl, cunit).to(default_cunit)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        flux = u.Quantity(flux, bunit).to(default_bunit, u.equivalencies.spectral_density(wl))
    return wl.value, flux.value


def readOSCspec(filepath):
    """
    Read spectra from a JSON file from the Open Supernova Catalog (https://sne.space)

    Parameters
    ----------
    filepath : str
        Path to the JSON file containing the spectra

    Returns
    -------
    filenames : list
        List of filenames corresponding to each spectrum
    times : list
        List of :class:`astropy.time.Time` when each spectrum was observed
    tel : list
        List of telescope names (if given) where each spectrum was observed
    inst : list
        List of instrument names (if given) with which each spectrum was observed
    wl : list
        List of wavelength arrays for each spectrum
    fx : list
        List of flux arrays for each spectrum
    scales : list
        List of scaling factors for each spectrum (all ones for the OSC)
    """
    with open(filepath) as f:
        json_dict = json.load(f)
    rows = json_dict[os.path.splitext(os.path.basename(filepath))[0]]
    if 'spectra' in rows:
        rows = rows['spectra']
    else:
        return [], [], [], [], [], [], []
    superdict = {key: [d[key] if key in d else '0' for d in rows] for key in
                 set(np.concatenate([list(d.keys()) for d in rows]))}
    times = [Time(float(t), format=u.lower()) for t, u in zip(superdict['time'], superdict['u_time'])]
    wl = [0.1 * np.array(d, dtype=float)[:, 0] for d in superdict['data']]
    fx = [np.array(d, dtype=float)[:, 1] for d in superdict['data']]
    if 'telescope' in superdict:
        tel = superdict['telescope']
    else:
        tel = np.tile('', len(rows))
    if 'instrument' in superdict:
        inst = superdict['instrument']
    else:
        inst = np.tile('', len(rows))
    #    t = Table([superdict['filename'], times, tel, inst, wl, fx, np.ones(len(rows))],
    #               names=['filename', 'date', 'telescope', 'instrument', 'wl', 'flux', 'scale'])
    return superdict['filename'], times, tel, inst, wl, fx, np.ones(len(rows))


def readspec(f, verbose=False, return_header=False):
    """
    Read a spectrum from a FITS or ASCII file and try to identify where and when it was observed

    Parameters
    ----------
    f : str
        Filename from which to read the spectrum
    verbose : bool, optional
        If True, print the date and filename of the spectrum
    return_header : bool, optional
        If True, return the full header as an additional return value

    Returns
    -------
    x : array-like
        Wavelengths, typically in ångströms
    y : array-like
        Fluxes in erg / (s cm2 angstrom), if units are identifiable
    date : astropy.time.Time
        Time at which the spectrum was observed, if identifiable (otherwise ``None``)
    telescope : str
        Name of the telescope at which the spectrum was observed, if identifiable (otherwise ``''``)
    instrument: str
        Name of the instrument used to observe the spectrum, if identifiable (otherwise ``''``)
    hdr : dict-like
        If ``return_header=True``, metadata is returned as a FITS header or dictionary (for ASCII files)
    """
    ext = os.path.splitext(f)[1]
    if ext == '.fits':
        x, y, hdr = readfitsspec(f, header=True)
    elif ext == '.json':
        x, y, hdr = readOSCspec(f)
    else:  # assume it's some ASCII format
        t = Table.read(f, format='ascii')
        # assume it's the first two columns, regardless of if column names are given
        x = t.columns[0]
        y = t.columns[1]
        hdr = {}
        comments = t.meta.get('comments', [])
        for line in comments:
            match = re.search('([^ ]*) *[=:] *([^/]*)', line)
            if match is None:
                continue
            kwd, val = match.groups()  # (anything before first '='/':', anything between first '='/':' and '/')
            hdr[kwd.strip(' #')] = val.strip(' "\'')
    for kwd in ['MJD-OBS', 'MJD_OBS', 'MJD', 'JD', 'DATE-AVG', 'UTMIDDLE', 'DATE-OBS', 'DATE_BEG', 'UTSHUT', 'OBS_DATE',
                'AVE_MJD']:
        if kwd in hdr and hdr[kwd]:
            if 'MJD' in kwd:
                date = Time(float(hdr[kwd]), format='mjd')
            elif 'JD' in kwd and float(hdr['JD']) > 2400000:
                date = Time(float(hdr[kwd]), format='jd')
            elif 'JD' in kwd:
                date = Time(float(hdr[kwd]) + 2400000, format='jd')
            elif 'T' in hdr[kwd]:
                date = Time(hdr[kwd])
            elif kwd == 'OBS_DATE':
                date = Time(hdr[kwd].split('+')[0])
            elif '-' in hdr[kwd]:
                for kwd2 in ['UTMIDDLE', 'EXPSTART', 'UT']:
                    if kwd2 in hdr and type(hdr[kwd2]) == str and ':' in hdr[kwd2]:
                        date = Time(hdr[kwd] + 'T' + hdr[kwd2])
                        break
                    elif kwd2 in hdr:
                        h = int(np.floor(hdr[kwd2]))
                        m = int(np.floor((hdr[kwd2] * 60) % 60))
                        s = int(np.floor((hdr[kwd2] * 3600) % 60))
                        date = Time(hdr[kwd] + 'T{:02d}:{:02d}:{:02d}'.format(h, m, s))
                        break
                else:
                    date = Time(hdr[kwd])
            else:
                continue
            break
    else:  # hope it's in the filename
        m1 = re.search('24[0-9][0-9][0-9][0-9][0-9]\.[0-9]+', f)  # JD w/1 or more decimals
        m2 = re.search('([12][90][0-9][0-9])-?(0[0-9]|1[0-2])-?(0[1-9]|[12][0-9]|3[01])', f)  # YYYYMMDD
        m_tns = re.search(
            '(19|20)[0-9][0-9]-(0[0-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])_([01][0-9]|2[0-4])-[0-5][0-9]-[0-5][0-9]',
            f)  # YYYY-MM-DD_HH:MM:SS
        m3 = re.search('[0-9][0-9][0-9]d', f)  # integer phase followed by 'd'
        m4 = re.search('[0-9][0-9][0-9][0-9][0-9](\.[0-9]+)?', f)  # MJD w/1 or more decimals
        if m1 is not None:
            m = m1.group()
            date = Time(float(m), format='jd')
        elif m2 is not None:
            date = Time('-'.join(m2.groups()))
        elif m_tns is not None:
            m = m_tns.group()
            d, t = m.split('_')
            date = Time(d + 'T' + t.replace('-', ':'))
        elif m3 is not None:
            m = m3.group()
            date = Time(float(m[:-1]), format='mjd')
        elif m4 is not None:
            m = m4.group()
            date = Time(float(m), format='mjd')
        else:
            date = None

    if 'TELESCOP' in hdr:
        telescope = hdr['TELESCOP'].strip()
    elif 'TELESCOPE' in hdr:
        telescope = hdr['TELESCOPE'].strip()
    elif 'OBSERVAT' in hdr:
        telescope = hdr['OBSERVAT'].strip()
    else:
        telescope = ''
    if 'INSTRUME' in hdr:
        instrument = hdr['INSTRUME'].strip()
    elif 'INSTRUMENT' in hdr:
        instrument = hdr['INSTRUMENT'].strip()
    elif 'INSTR' in hdr:
        instrument = hdr['INSTR'].strip()
    elif 'INSTRUMENT_ID' in hdr:
        instrument = hdr['INSTRUMENT_ID']
    else:
        instrument = ''

    x, y = convert_spectrum_units(x, y, hdr)

    if verbose:
        print(date.isot, f)
    if return_header:
        return x, y, date, telescope, instrument, hdr
    else:
        return x, y, date, telescope, instrument


def calibrate_spectra(spectra, lc, filters=None, order=0, subtract_percentile=None, max_extrapolate=1., show=False):
    """
    Calibrate a set of spectra to an observed broadband light curve.

    Each calibrated spectrum is saved to a text file that corresponds to the original filename prefixed by ``photcal_``.

    Parameters
    ----------
    spectra : list
        List of filenames containing the spectra
    lc : lightcurve_fitting.lightcurve.LC
        Photometry table containing the observed light curve
    filters : list, optional
        Only use this subset of filters for calibration
    order : int, optional
        Polynomial order for the calibration function. Default: 0 (constant factor)
    subtract_percentile : float, optional
        Subtract flux corresponding to this percentile of the spectrum before calibration. Default: no subtraction
    max_extrapolate : float, optional
        Assume constant flux in a filter for this many days after the last observed point. Default: 1 day.
    show : bool, optional
        Plot the observed light curve and the uncalibrated and calibrated spectra, and ask whether to save the results
    """
    if filters is not None:
        lc = lc.where(filter=[filtdict[f] for f in filters])
    lc.calcFlux()
    lc.sort('MJD')
    filts = set(lc['filter'])

    for filt in filts:
        filt.read_curve()
        filt.trans.sort('freq')

    plt.ion()
    fig = plt.figure(figsize=(8., 6.))

    for spec in spectra:
        wl, flux, time, _, _ = readspec(spec)
        mjd = time.mjd
        if show:
            fig.clf()
            ax1 = plt.subplot(211)
            lc.plot(xcol='MJD', ycol='flux', offset_factor=0)
            ax1.axvline(mjd)
            ax1.set_xlabel('MJD')
            ax1.set_ylabel('$F_\\nu$ (W Hz$^{-1}$)')
            ax2 = plt.subplot(212)
        good = ~np.isnan(flux)
        lam = wl[good] * u.angstrom
        Flam = flux[good] * u.erg / u.s / u.angstrom / u.cm**2
        nu = const.c / lam
        Fnu = (Flam * lam / nu).to(u.W / u.Hz / u.m**2).value[::-1]
        nu = nu.to(u.THz).value[::-1]
        if subtract_percentile is not None:
            Fnu -= np.nanpercentile(Fnu, subtract_percentile)
        freqs = []
        ratios = []
        for filt in filts:
            freq0 = filt.freq_eff.value - filt.freq_range[0]
            freq1 = filt.freq_range[1] + filt.freq_eff.value
            if freq1 < np.min(nu) or freq0 > np.max(nu):
                print(filt, "and spectrum don't overlap")
                continue  # filter and spectrum don't overlap
            lc_filt = lc.where(filter=filt, nondet=False)
            if len(lc_filt) == 0 or mjd - np.max(lc_filt['MJD']) > max_extrapolate or mjd < np.min(lc_filt['MJD']):
                print(filt, "not observed before and after spectrum")
                continue
            flux_lc = np.interp(mjd, lc_filt['MJD'], lc_filt['flux'])
            trans_interp = np.interp(nu, filt.trans['freq'], filt.trans['T_norm_per_freq'])
            flux_spec = np.trapz(Fnu * trans_interp, nu) / np.trapz(trans_interp, nu)
            ratio = flux_lc / flux_spec
            if show:
                ax2.axvspan(freq0, freq1, color=filt.color, alpha=0.2)
                ax2.plot(filt.freq_eff, flux_lc, marker='o', zorder=5, **filt.plotstyle)
            ratios.append(ratio)
            freqs.append(filt.freq_eff.value)
        if not ratios:
            print('no filters for', spec)
            if show:
                plt.close(fig)
            continue
        scale = np.mean(ratios)
        if order:
            p = np.polyfit(freqs, np.array(ratios) / scale, order)
            corr = np.polyval(p, nu) * scale
            print(spec, scale, p[:-1])
        else:
            corr = scale
            print(spec, scale)
        if show:
            ax2.plot(nu, Fnu * scale, label='rescaled')
            ax2.set_xlabel('Frequency (THz)')
            ax2.set_ylabel('$F_\\nu$ (W Hz$^{-1}$)')
            if order:
                ax2.plot(nu, Fnu * corr, color='C2', label='rescaled & warped')
                plt.legend(loc='best')
            plt.pause(0.1)
            ans = input('accept this scale? [Y/n] ')
        if not show or ans.lower() != 'n':
            data_out = np.array([wl[good], flux[good] * corr]).T
            path_in, filename_in = os.path.split(spec)
            filename_out = os.path.join(path_in, 'photcal_' + filename_in).replace('.fits', '.txt')
            np.savetxt(filename_out, data_out, fmt='%.1f %.2e')
            print(filename_out)


def create_wiserep_tsv(specpaths, wiserep_dir, verbose=False):
    """
    Prepares a TSV file for uploading spectra to WISeREP (see https://www.wiserep.org/content/wiserep-getting-started).

    Also copies all the spectrum files to a single directory, and converts any FITS files to ASCII.

    Parameters
    ----------
    specpaths: list
        List of paths to the spectrum files. Optionally, data quality can be specified in a tuple, e.g.,
        [(filename1.fits, 3), (filename2.txt, 2), ...], where the second element is 1 (low), 2 (medium), or 3 (high).
    wiserep_dir: str
        Directory in which to collect the spectra. If it already exists, it will be deleted and remade.
        The TSV file will be saved alongside this directory using the same name but with a '.tsv' extension.
    verbose: bool, optional
        If True, print a message when any file is copied or created.
    """

    # prepare the output directory
    if os.path.exists(wiserep_dir):
        ans = input(f'Are you sure you want to delete the directory {wiserep_dir}? [y/N] ')
        if ans.lower() != 'y':
            return
        shutil.rmtree(wiserep_dir)
    os.mkdir(wiserep_dir)

    # collect the metadata
    bibcode = input('bibcode: ')
    rows = []
    instruments = {}
    for specpath in specpaths:
        if isinstance(specpath, tuple):
            specpath, quality = specpath
            quality = min(max(round(quality), 1), 3)  # forces it to be 1, 2, or 3
        else:
            quality = 2  # medium quality
        specfile = os.path.split(specpath)[-1]
        ascii_file = specfile.replace('.fits', '.txt').replace('.csv', '.txt')
        print()  # empty line between spectra
        wl, flux, date, tel, inst, hdr = readspec(specpath, verbose=True, return_header=True)
        groups = input('https://www.wiserep.org/groups\ngroup IDs (comma sep.): ')
        if inst not in instruments:
            inst_id = input(f'https://www.wiserep.org/aux\nlook up instrument ID for {inst} (required): ')
            if inst and inst_id:
                instruments[inst] = int(inst_id)
        row = [
            ascii_file,
            specfile if specfile.endswith('.fits') else None,
            date.iso,
            instruments.get(inst, inst_id),
            hdr.get('exptime'),
            {'angstrom': 11, 'nm': 12, 'um': 13}.get(hdr.get('CUNIT1', hdr.get('XUNITS', 'angstrom')).lower()),
            1,  # wavelength in air, hardcoded for now
            1.,  # erg/cm2/s/Å, hardcoded for now
            6,  # erg/cm2/s/Å, hardcoded for now
            2 if specfile.startswith('photcal') else 1,  # method of flux calibration (1 = std star, 2 = photometry)
            0,  # not extinction corrected
            hdr.get('OBSERVER', 'Unknown'),
            hdr.get('REDUCER'),
            None,  # reduction date
            float(hdr['APERWID']) if 'APERWID' in hdr else None,
            hdr.get('DICHROIC'),
            hdr.get('GRISM'),
            hdr.get('GRATING'),
            float(hdr['BLAZE']) if 'BLAZE' in hdr else None,
            float(hdr['AIRMASS']) if 'AIRMASS' in hdr else None,
            hdr.get('HA') or None,
            10,  # spectrum of object
            quality,  # spectrum quality
            0.,  # spectrum proprietary period
            'days',  # prop. period units
            groups,
            None,  # comments
            bibcode or None,
            None,  # contributor
            None,  # related files
            None,  # related files
            None,  # related files
            None,  # related files
        ]
        rows.append(row)

        # copy and convert files
        if not specfile.endswith('.csv'):
            shutil.copy(specpath, wiserep_dir)
            if verbose:
                print(f'copied {specfile} to {wiserep_dir}')
        if specfile.endswith('.fits') or specfile.endswith('.csv'):
            data_out = np.transpose([wl, flux])
            np.savetxt(os.path.join(wiserep_dir, ascii_file), data_out, fmt=('%f', '%e'), header=hdr.__repr__())
            if verbose:
                print(f'wrote {wiserep_dir}/{ascii_file}')

    # create the table and write the TSV file
    t = Table(rows=rows, names=[
        'Ascii-filename*',
        'FITS-filename*',
        'Obs-date* [YYYY-MM-DD HH:MM:SS] / JD',
        'Instrument-Id*',
        'Exp-time (sec)',
        'WL Units-id',
        'WL Medium-Id',
        'Flux Unit Coeff',
        'Flux Units-Id',
        'Flux Calib. By-Id',
        'Extinction-Corrected-Id',
        'Observer/s      ',
        'Reducer/s   ',
        'Reduction-date [YYYY-MM-DD HH:MM:SS] / JD',
        'Aperture (Slit)',
        'Dichroic',
        'Grism',
        'Grating',
        'Blaze',
        'Airmass',
        'Hour Angle',
        'Spec Type-Id',
        'Spec Quality-Id',
        'Spec. Prop-period value',
        'Prop-period units',
        'Assoc. Groups',
        'Spec-Remarks',
        'Publish (bibcode)',
        'Contrib',
        'Related-file1',
        'RF1 Comments',
        'Related-file2',
        'RF2 Comments'
    ], meta={'comments': ['TSV-type:\tspectra']})
    # use the lower-level interface to add the second line with the defaults
    writer = ascii.get_writer(ascii.Tab, fast_writer=False, comment='',
                              fill_values=[('None', 'NULL'), ('', 'NULL'), ('UNKNOWN', 'NULL')])
    lines = writer.write(t)
    lines.insert(2, '\t\t\t\tNULL\t[default=11 (Angstrom)]\t[default=1 (Air)]\t[default=1.0]\t[default=6]\tNULL\tNULL\t[Unknown]\tNULL\tNULL\tNULL\tNULL\tNULL\tNULL\tNULL\tNULL\tNULL\t[default=10=Object]\tNULL\tNULL\t[days/months/years]\t[Comma delim.]\tNULL\tNULL\tNULL\tNULL\tNULL\tNULL\tNULL')
    with open(wiserep_dir + '.tsv', 'w') as f:
        f.write('\n'.join(lines) + '\n')
    if verbose:
        print(f'\nwrote {wiserep_dir}.tsv')
    return t


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Calibrate spectra of a supernova to photometry.')
    parser.add_argument('spectra', nargs='+', help='filenames of spectra')
    parser.add_argument('--lc', help='filename of photometry table (must have columns "MJD", "filt", "mag"/"flux", and'
                                     '"dmag"/"dflux")')
    parser.add_argument('--lc-format', default='ascii',
                        help='format of photometry table (passed to :func:`astropy.table.Table.read`)')
    parser.add_argument('-f', '--filters', nargs='+', help='filters to use for calibration')
    parser.add_argument('-o', '--order', type=int, default=0, help='polynomial order of correction function')
    parser.add_argument('--subtract-percentile', type=float, help='subtract continuum from spectrum before correcting')
    parser.add_argument('--max-extrapolate', type=float,
                        help='assume constant flux in a filter for this many days after the last observed point')
    parser.add_argument('--show', action='store_true')
    args = parser.parse_args()

    lc = LC.read(args.lc, format=args.format)
    calibrate_spectra(args.spectra, lc, args.filters, args.order, args.subtract_percentile, args.max_extrapolate,
                      args.show)



import numpy as np
import json
from glob import glob

for filename in glob('*'):
    print(filename)
    with open(filename) as f:
        data = json.load(f)
    np.savetxt(data[0][0] + '.txt', data[1:], '%.1f %.9f')



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.path import Path
from astropy.table import Table, vstack, MaskedColumn
from .filters import filtdict
import itertools
from matplotlib.markers import MarkerStyle
try:
    from config import markers
except ModuleNotFoundError:
    markers = {}


class Arrow(Path):
    """
    An downward-pointing arrow-shaped ``Path``, whose head has half-width ``hx`` and height ``hy`` (in units of length)
    """
    def __init__(self, hx, hy):
        verts = [(0, 0),
                 (0, -1),
                 (-hx, -1 + hy),
                 (0, -1),
                 (hx, -1 + hy),
                 (0, -1),
                 (0, 0)]
        codes = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]
        Path.__init__(self, verts, codes)


arrow = Arrow(0.2, 0.3)
othermarkers = ('o', *MarkerStyle.filled_markers[2:])
itermarkers = itertools.cycle(othermarkers)
usedmarkers = []
itercolors = itertools.cycle(plt.rcParams['axes.prop_cycle'].by_key()['color'])

# if you edit this list, also add the new names to usage.rst
column_names = {
    'Filter': ['filt', 'filter', 'Filter', 'band', 'FLT', 'Band'],
    'Telescope': ['telescope', 'Telescope', 'Tel', 'tel+inst'],
    'Source': ['source', 'Source'],
    'Apparent Magnitude': ['mag', 'Magnitude', 'Mag', 'ab_mag', 'PSFmag', 'MAG', 'omag', 'magnitude', 'apparent_mag'],
    'Apparent Magnitude Uncertainty': [
        'dmag', 'Magnitude_Error', 'magerr', 'MagErr', 'mag_err', 'e_mag', 'Error', 'err', 'PSFerr', 'MAGERR', 'e_omag',
        'e_magnitude', 'apparent_mag_err', 'Mag_Err', 'emag',
    ],
    'MJD': ['MJD', 'mjd'],
    'JD': ['JD', 'jd'],
    'Phase (rest days)': ['phase', 'Phase', 'PHASE'],
    'Flux $F_ν$ (W m$^{-2}$ Hz$^{-1}$)': ['flux', 'FLUXCAL'],
    'Flux Uncertainty': ['dflux', 'FLUXCALERR'],
    'Nondetection': ['nondet', 'Is_Limit', 'UL', 'l_omag', 'upper_limit', 'upperlimit'],
    'Absolute Magnitude': ['absmag'],
    'Luminosity $L_ν$ (W Hz$^{-1}$)': ['lum'],
    'Luminosity Uncertainty': ['dlum'],
    'Effective Wavelength (nm)': ['wl_eff'],  # calculated from filters; does not need to be in usage.rst
}


class LC(Table):
    """
    A broadband light curve, stored as an :class:`astropy.table.Table`

    Attributes
    ----------
    sn : object
        Astronomical transient with which this light curve is associated
    nondetSigmas : float
        Significance level implied by nondetections in the light curve. Default: 3σ
    groupby : set
        Column names to group by when binning the light curve. Default: ``{'filt', 'filter', 'source'}``
    """
    def __init__(self, *args, **kwargs):
        Table.__init__(self, *args, **kwargs)
        self.sn = None
        self.nondetSigmas = 3.
        self.groupby = {'filt', 'filter', 'source'}
        self.markers = markers.copy()

    def where(self, **kwargs):
        """
        Select the subset of a light curve matching some criteria, given as keyword arguments, e.g., ``colname=value``.

        The keyword ``colname`` can be any of the following:
          * a column in the table, in which case rows must match ``value`` in that column
          * a column in the table + ``_not``, in which case rows must *not* match ``value`` in that column
          * a column in the table + ``_min``, in which case rows must be >= ``value`` in that column
          * a column in the table + ``_max``, in which case rows must be <= ``value`` in that column

        ``value`` must match the data type of the column ``colname`` and can either be a single value or a list of
        values. If ``value`` is a list, rows must match at least one of the values. If ``value`` is a list and
        ``colname`` ends in ``_not``, rows must not match any of the values.
        """
        use = np.tile(True, len(self))
        for col, val in kwargs.items():
            if isinstance(val, list):
                if '_not' in col:
                    use1 = np.tile(True, len(self))
                    for v in val:
                        use1 &= self[col.replace('_not', '')] != v
                else:
                    use1 = np.tile(False, len(self))
                    for v in val:
                        use1 |= self[col] == v
            elif '_min' in col:
                use1 = self[col.replace('_min', '')] >= val
            elif '_max' in col:
                use1 = self[col.replace('_max', '')] <= val
            elif '_not' in col:
                if val is None:
                    use1 = np.array([v is not None for v in self[col.replace('_not', '')]])
                else:
                    use1 = self[col.replace('_not', '')] != val
            else:
                if val is None:
                    use1 = np.array([v is None for v in self[col]])
                else:
                    use1 = self[col] == val
            use &= use1
        selected = self[use]
        selected.sn = self.sn
        selected.meta = self.meta
        selected.markers = self.markers
        return selected

    def get(self, key, default=None):
        return self[key] if key in self.colnames else default

    def normalize_column_names(self):
        """
        Rename any recognizable columns to their standard names for this package (see `lightcurve.column_names`).
        """
        for good_key, *bad_keys in column_names.values():
            if good_key not in self.colnames:
                for bad_key in bad_keys:
                    if bad_key in self.colnames:
                        self.rename_column(bad_key, good_key)
                        break
        if 'MJD' not in self.colnames and 'JD' in self.colnames:
            self['MJD'] = self['JD'] - 2400000.5
            self.remove_column('JD')
        if 'nondet' in self.colnames and self['nondet'].dtype != bool:
            if isinstance(self['nondet'], MaskedColumn):
                self['nondet'] = self['nondet'].filled()
            nondet = (self['nondet'] == 'True') | (self['nondet'] == 'T') | (self['nondet'] == '>')
            self.remove_column('nondet')
            self['nondet'] = nondet

    def filters_to_objects(self, read_curve=True):
        """
        Parse the ``'filt'`` column into :class:`filters.Filter` objects and store in the ``'filter'`` column

        Parameters
        ----------
        read_curve : bool, optional
            Read in the transmission function for each filter encountered (default)
        """
        self['filter'] = [filtdict['?'] if np.ma.is_masked(f) or str(f) not in filtdict else filtdict[str(f)]
                          for f in self['filt']]
        is_swift = np.zeros(len(self), bool)
        if 'telescope' in self.colnames:
            is_swift |= self['telescope'] == 'Swift'
            is_swift |= self['telescope'] == 'UVOT'
            is_swift |= self['telescope'] == 'Swift/UVOT'
            is_swift |= self['telescope'] == 'Swift+UVOT'
        if 'source' in self.colnames:
            is_swift |= self['source'] == 'SOUSA'
        if is_swift.any():
            for filt, swiftfilt in zip('UBV', 'sbv'):
                self['filter'][is_swift & (self['filt'] == filt)] = filtdict[swiftfilt]
        if read_curve:
            for filt in np.unique(self['filter']):
                filt.read_curve()

    @property
    def zp(self):
        """
        Returns an array of zero points for each filter in the ``'filter'`` column
        """
        return np.array([f.m0 for f in self['filter']])

    def calcFlux(self, nondetSigmas=None, zp=None):
        """
        Calculate the ``'flux'`` and ``'dflux'`` columns from the ``'mag'`` and ``'dmag'`` columns

        Parameters
        ----------
        nondetSigmas : float, optional
            Significance level implied by nondetections in the light curve. Default: 3σ
        zp : float, array-like, optional
            Array of zero points for each magnitude. Default: standard for each filter
        """
        if nondetSigmas is not None:
            self.nondetSigmas = nondetSigmas
        if zp is None:
            zp = self.zp
        self['flux'], self['dflux'] = mag2flux(self['mag'], self['dmag'], zp, self.get('nondet'), self.nondetSigmas)

    def bin(self, delta=0.3, groupby=None):
        """
        Bin the light curve by averaging points within ``delta`` days of each other

        Parameters
        ----------
        delta : float, optional
            Bin size, in days. Default: 0.3 days
        groupby : set, optional
            Column names to group by before binning. Default: ``{'filt', 'filter', 'source'}``

        Returns
        -------
        lc : lightcurve_fitting.lightcurve.LC
            Binned light curve
        """
        if groupby is not None:
            self.groupby = groupby
        subtabs = []
        self.groupby = list(set(self.groupby) & set(self.colnames))
        if self.groupby:
            grouped = self.group_by(self.groupby)
        else:
            grouped = self
        for g, k in zip(grouped.groups, grouped.groups.keys):
            mjd, flux, dflux = binflux(g['MJD'], g['flux'], g['dflux'], delta)
            binned = LC([mjd, flux, dflux], names=['MJD', 'flux', 'dflux'])
            for key in self.groupby:
                binned[key] = k[key]
            subtabs.append(binned)
        lc = vstack(subtabs)
        lc.sn = self.sn
        lc.meta = self.meta
        return lc

    def findNondet(self, nondetSigmas=None):
        """
        Add a boolean column ``'nondet'`` indicating flux measurements that are below the detection threshold

        Parameters
        ----------
        nondetSigmas : float, optional
            Significance level implied by nondetections in the light curve. Default: 3σ
        """
        if nondetSigmas is not None:
            self.nondetSigmas = nondetSigmas
        self['nondet'] = self['flux'] < self.nondetSigmas * self['dflux']

    def calcMag(self, nondetSigmas=None, zp=None):
        """
        Calculate the ``'mag'`` and ``'dmag'`` columns from the ``'flux'`` and ``'dflux'`` columns

        Parameters
        ----------
        nondetSigmas : float, optional
            Significance level implied by nondetections in the light curve. Default: 3σ
        zp : float, array-like, optional
            Array of zero points for each magnitude. Default: standard for each filter
        """
        if nondetSigmas is not None:
            self.nondetSigmas = nondetSigmas
        self.findNondet()
        if zp is None:
            zp = self.zp
        self['mag'], self['dmag'] = flux2mag(self['flux'], self['dflux'], zp, self.get('nondet'), self.nondetSigmas)

    def calcAbsMag(self, dm=None, extinction=None, hostext=None):
        """
        Calculate the ``'absmag'`` column from the ``'mag'`` column by correcting for distance and extinction

        Parameters
        ----------
        dm : float, optional
            Distance modulus. Default: use the distance modulus of ``self.sn``, if any. Otherwise do not correct for
            distance.
        extinction : dict, optional
            Milky Way extinction coefficients :math:`A_λ` for each filter. Default: use the extinction of ``self.sn``,
            if any. Otherwise do not correct for Milky Way extinction.
        hostext : dict, optional
            Host galaxy extinction coefficients :math:`A_λ` for each filter. Default: use the extinction of ``self.sn``,
            if any. Otherwise do not correct for host galaxy extinction.
        """
        if dm is not None:
            self.meta['dm'] = dm
        elif self.sn is not None:
            self.meta['dm'] = self.sn.dm
        elif 'dm' not in self.meta:
            self.meta['dm'] = 0.

        if extinction is not None:
            self.meta['extinction'] = extinction
        elif self.sn is not None:
            self.meta['extinction'] = self.sn.extinction
        elif 'extinction' not in self.meta:
            self.meta['extinction'] = {}

        if hostext is not None:
            self.meta['hostext'] = hostext
        elif self.sn is not None:
            self.meta['hostext'] = self.sn.hostext
        elif 'hostext' not in self.meta:
            self.meta['hostext'] = {}

        self['absmag'] = self['mag'].data - self.meta['dm']
        for filtobj in set(self['filter']):
            for filt in filtobj.names:
                if filt in self.meta['extinction']:
                    self['absmag'][self['filter'] == filtobj] -= self.meta['extinction'][filt]
                    break
            else:
                print('MW extinction not applied to filter', filtobj)
            for filt in filtobj.names:
                if filt in self.meta['hostext']:
                    self['absmag'][self['filter'] == filtobj] -= self.meta['hostext'][filt]
                    break
            else:
                print('host extinction not applied to filter', filtobj)

    def calcLum(self, nondetSigmas=None):
        """
        Calculate the ``'lum'`` and ``'dlum'`` columns from the ``'absmag'`` and ``'dmag'`` columns

        Parameters
        ----------
        nondetSigmas : float, optional
            Significance level implied by nondetections in the light curve. Default: 3σ
        """
        if nondetSigmas is not None:
            self.nondetSigmas = nondetSigmas
        self['lum'], self['dlum'] = mag2flux(self['absmag'], self['dmag'], self.zp + 90.19, self.get('nondet'),
                                             self.nondetSigmas)

    def findPeak(self, **criteria):
        """
        Find the peak of the light curve and store it in ``self.sn.peakdate``

        Parameters
        ----------
        criteria : dict, optional
            Use only a subset of the light curve matching some criteria when calculating the peak date (stored in
            ``self.sn.peakcriteria``
        """
        if 'nondet' in self.colnames:
            criteria['nondet'] = False
        peaktable = self.where(**criteria)
        if len(peaktable):
            imin = np.argmin(peaktable['mag'])
            self.meta['peakdate'] = peaktable['MJD'][imin]
        else:
            self.meta['peakdate'] = np.nan
        if self.sn is not None:
            self.meta['redshift'] = self.sn.z  # needed for calcPhase
            self.sn.peakdate = self.meta['peakdate']
            self.sn.peakcriteria = criteria

    def calcPhase(self, rdsp=False, hours=False):
        """
        Calculate the rest-frame ``'phase'`` column from the ``'MJD'`` column and ``self.sn.refmjd`` and ``self.sn.z``

        Parameters
        ----------
        rdsp : bool, optional
            Define phase as rest-frame days since peak, rather than rest-frame days since explosion
        """
        if 'refmjd' not in self.meta and self.sn is not None:
            if rdsp and self.sn.peakdate is None:
                raise Exception('must run sn.findPeak() first')
            elif rdsp:
                self.sn.refmjd = self.sn.peakdate
            elif self.sn.explosion is not None:
                self.sn.refmjd = self.sn.explosion
            else:
                if 'nondet' in self.colnames:
                    detections = self.where(nondet=False)
                else:
                    detections = self
                self.sn.refmjd = np.min(detections['MJD'].data)
            self.meta['refmjd'] = self.sn.refmjd
            self.meta['redshift'] = self.sn.z
        self['phase'] = (self['MJD'].data - self.meta['refmjd']) / (1 + self.meta['redshift'])
        if 'dMJD0' in self.colnames:
            self['dphase0'] = self['dMJD0'] / (1. + self.meta['redshift'])
        if 'dMJD1' in self.colnames:
            self['dphase1'] = self['dMJD1'] / (1. + self.meta['redshift'])
        if hours:
            self['phase'] *= 24.
            if 'dphase0' in self.colnames:
                self['dphase0'] *= 24.
            if 'dphase1' in self.colnames:
                self['dphase1'] *= 24.

    def plot(self, xcol='phase', ycol='absmag', offset_factor=1., color='filter', marker=None, use_lines=False,
             normalize=False, fillmark=True, **kwargs):
        """
        Plot the light curve, with nondetections marked with a downward-pointing arrow

        Parameters
        ----------
        xcol : str, optional
            Column to plot on the horizontal axis. Default: ``'phase'``
        ycol : str, optional
            Column to plot on the vertical axis. Default: ``'absmag'``
        offset_factor : float, optional
            Increase or decrease the filter offsets by a constant factor. Default: 1.
        color : str, optional
            Column that controls the color of the lines and points. Default: ``'filter'``
        marker : str, optional
            Column that controls the marker shape. Default: ``'source'`` or ``'telescope'``
        use_lines : bool, optional
            Connect light curve points with lines. Default: False
        normalize : bool, optional
            Normalize all light curves to peak at 0. Default: False
        fillmark : bool, optional
            Fill each marker with color. Default: True
        kwargs : dict, optional
            Keyword arguments matching column names in the light curve are used to specify a subset of points to plot.
            Additional keyword arguments passed to :func:`matplotlib.pyplot.plot`.
        """
        if xcol.startswith('filter'):
            unit = xcol.split(':')[-1] if ':' in xcol else None
            xcol = 'wl_eff'
            self[xcol] = [f.wl_eff.to(unit) if unit else f.wl_eff for f in self['filter']]
        xchoices = ['phase', 'MJD']
        while xcol not in self.keys():
            xchoices.remove(xcol)
            if xchoices:
                xcol = xchoices[0]
            else:
                raise Exception('no columns found for x-axis')
        ychoices = ['absmag', 'mag']
        while ycol not in self.keys():
            ychoices.remove(ycol)
            if ychoices:
                ycol = ychoices[0]
            else:
                raise Exception('no columns found for y-axis')
        if marker is None:
            if 'source' in self.colnames:
                marker = 'source'
            elif 'telescope' in self.colnames:
                marker = 'telescope'
            else:
                marker = 'o'
        criteria = {key: val for key, val in kwargs.items() if key in self.colnames}
        plot_kwargs = {key: val for key, val in kwargs.items() if key not in self.colnames}
        plottable = self.where(**criteria)
        if len(plottable) == 0:
            return
        groupby = set()
        if color in plottable.keys():
            groupby.add(color)
        if marker in plottable.keys():
            groupby.add(marker)
        if groupby:
            plottable = plottable.group_by(list(groupby))
            keys = plottable.groups.keys
        else:
            keys = [Table()]
        if self.sn is None:
            linestyle = plot_kwargs.pop('linestyle') if 'linestyle' in plot_kwargs else None
            linewidth = plot_kwargs.pop('linewidth') if 'linewidth' in plot_kwargs else None
        else:
            linestyle = self.sn.linestyle
            linewidth = self.sn.linewidth
        for g, k in zip(plottable.groups, keys):
            filt = g['filter'][0]
            if color == 'filter':
                col = filt.color
                mec = 'k' if filt.system == 'Johnson' else filt.linecolor
            elif color == 'name':
                col = self.sn.plotcolor
                mec = col if col not in ['w', '#FFFFFF'] else 'k'
            else:
                col = mec = next(itercolors)
            mfc = col if fillmark else 'none'
            if marker == 'name':
                mark = self.sn.marker
            elif marker in plottable.keys():
                if g[marker][0] not in self.markers:
                    for nextmarker in othermarkers:
                        if nextmarker not in usedmarkers:
                            self.markers[g[marker][0]] = nextmarker
                            break
                    else:
                        self.markers[g[marker][0]] = next(itermarkers)
                mark = self.markers[g[marker][0]]
            elif marker in MarkerStyle.markers:
                mark = marker
            elif marker == 'none':
                mark = None
            else:
                mark = next(itermarkers)
            usedmarkers.append(mark)
            if use_lines:
                g.sort(xcol)
            elif 'mag' in ycol:
                yerr = g['dmag']
            else:
                yerr = g['d' + ycol]
            x = g[xcol].data
            y = g[ycol].data - filt.offset * offset_factor
            if normalize and ycol == 'mag':
                y -= self.sn.peakmag
            elif normalize and ycol == 'absmag':
                y -= self.sn.peakabsmag
            if 'mag' in ycol and 'nondet' in g.keys() and marker:  # don't plot if no markers used
                plt.plot(x[g['nondet']], y[g['nondet']], marker=arrow, linestyle='none', ms=25, mec=mec, **plot_kwargs)
            if 'filter' in k.colnames:
                if len(filt.name) >= 4 and not filt.offset:
                    k['filter'] = filt.name
                elif offset_factor:
                    k['filter'] = '${}{:+.0f}$'.format(filt.name, -filt.offset * offset_factor)
                else:
                    k['filter'] = '${}$'.format(filt.name)
            label = ' '.join([str(kv) for kv in k.values()])
            if not use_lines:
                plt.errorbar(x, y, yerr, color=mec, mfc=mfc, mec=mec, marker=mark, linestyle='none', label=label,
                             **plot_kwargs)
            elif 'mag' in ycol and 'nondet' in g.colnames:
                plt.plot(x[~g['nondet']], y[~g['nondet']], color=col, mfc=mfc, mec=mec, marker=mark, label=label,
                         linestyle=linestyle, linewidth=linewidth, **plot_kwargs)
                plt.plot(x[g['nondet']], y[g['nondet']], color=mec, mfc=mfc, mec=mec, marker=mark, linestyle='none',
                         **plot_kwargs)
            else:
                plt.plot(x, y, color=col, mfc=mfc, mec=mec, marker=mark, label=label, linestyle=linestyle,
                         linewidth=linewidth, **plot_kwargs)
        ymin, ymax = plt.ylim()
        if 'mag' in ycol and ymax > ymin:
            plt.ylim(ymax, ymin)
        for axlabel, keys in column_names.items():
            if xcol in keys:
                plt.xlabel(axlabel)
            elif ycol in keys:
                plt.ylabel(axlabel)

    @classmethod
    def read(cls, filepath, format='ascii', fill_values=None, **kwargs):
        if fill_values is None:
            fill_values = [('--', '0'), ('', '0')]
        t = super(LC, cls).read(filepath, format=format, fill_values=fill_values, **kwargs)
        t.normalize_column_names()
        if 'filt' in t.colnames:
            t.filters_to_objects()
        return t


def flux2mag(flux, dflux=np.array(np.nan), zp=0., nondet=None, nondetSigmas=3.):
    """
    Convert flux (and uncertainty) to magnitude (and uncertainty). Nondetections are converted to limiting magnitudes.

    Parameters
    ----------
    flux : float, array-like
        Flux to be converted
    dflux : float, array-like, optional
        Uncertainty on the flux to be converted. Default: :mod:`numpy.nan`
    zp : float, array-like, optional
        Zero point to be applied to the magnitudes
    nondet : array-like
        Boolean or index array indicating the nondetections among the fluxes. Default: no nondetections
    nondetSigmas : float, optional
        Significance level implied by nondetections in the light curve. Default: 3σ

    Returns
    -------
    mag : float, array-like
        Magnitude corresponding to the input flux
    dmag : float, array-like
        Uncertainty on the output magnitude
    """
    flux = flux.copy()
    dflux = dflux.copy()
    if nondet is not None:
        flux[nondet] = nondetSigmas * dflux[nondet]
        dflux[nondet] = np.nan
    mag = -2.5 * np.log10(flux) + zp
    dmag = 2.5 * dflux / (flux * np.log(10))
    return mag, dmag


def mag2flux(mag, dmag=np.nan, zp=0., nondet=None, nondetSigmas=3.):
    """
    Convert magnitude (and uncertainty) to flux (and uncertainty). Nondetections are assumed to imply zero flux.

    Parameters
    ----------
    mag : float, array-like
        Magnitude to be converted
    dmag : float, array-like, optional
        Uncertainty on the magnitude to be converted. Default: :mod:`numpy.nan`
    zp : float, array-like, optional
        Zero point to be applied to the magnitudes
    nondet : array-like
        Boolean or index array indicating the nondetections among the fluxes. Default: no nondetections
    nondetSigmas : float, optional
        Significance level implied by nondetections in the light curve. Default: 3σ

    Returns
    -------
    flux : float, array-like
        Flux corresponding to the input magnitude
    dflux : float, array-like
        Uncertainty on the output flux
    """
    flux = 10 ** ((zp - mag) / 2.5)
    dflux = np.log(10) / 2.5 * flux * dmag
    if nondet is not None:
        dflux[nondet] = flux[nondet] / nondetSigmas
        flux[nondet] = 0
    return flux, dflux


def binflux(time, flux, dflux, delta=0.2, include_zero=True):
    """
    Bin a light curve by averaging points within ``delta`` of each other in time

    Parameters
    ----------
    time, flux, dflux : array-like
        Arrays of times, fluxes, and uncertainties comprising the observed light curve
    delta : float, optional
        Bin size, in the same units as ``time``. Default: 0.2
    include_zero : bool, optional
        Include data points with no error bar

    Returns
    -------
    time, flux, dflux : array-like
        Binned arrays of times, fluxes, and uncertainties
    """
    bin_time = []
    bin_flux = []
    bin_dflux = []
    while len(flux) > 0:
        grp = np.array(abs(time - time[0]) <= delta)
        time_grp = time[grp]
        flux_grp = flux[grp]
        dflux_grp = dflux[grp]

        # Indices with no error bar
        zeros = (dflux_grp == 0) | (dflux_grp == 999) | (dflux_grp == 9999) | (dflux_grp == -1) | np.isnan(dflux_grp)
        if np.ma.is_masked(dflux_grp):
            zeros = zeros.data | dflux_grp.mask

        if any(zeros) and include_zero:
            x = np.mean(time_grp)
            y = np.mean(flux_grp)
            z = 0.
        else:
            # Remove points with no error bars
            time_grp = time_grp[~zeros]
            flux_grp = flux_grp[~zeros]
            dflux_grp = dflux_grp[~zeros]

            x = np.mean(time_grp)
            y = np.sum(flux_grp * dflux_grp ** -2) / np.sum(dflux_grp ** -2)
            z = np.sum(dflux_grp ** -2) ** -0.5
        # Append result 
        bin_time.append(x)
        bin_flux.append(y)
        bin_dflux.append(z)
        # Remove data points already used
        time = time[~grp]
        flux = flux[~grp]
        dflux = dflux[~grp]
    time = np.array(bin_time)
    flux = np.array(bin_flux)
    dflux = np.array(bin_dflux)
    return time, flux, dflux



from os import path
from setuptools import setup, find_packages
import sys
import versioneer


# NOTE: This file must remain Python 2 compatible for the foreseeable future,
# to ensure that we error out properly for people with outdated setuptools
# and/or pip.
min_version = (3, 6)
if sys.version_info < min_version:
    error = """
lightcurve_fitting does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(*(sys.version_info[:2] + min_version))
    sys.exit(error)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'requirements.txt')) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [line for line in requirements_file.read().splitlines()
                    if not line.startswith('#')]


setup(
    name='lightcurve-fitting',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Tools to fit analytical models to multiband light curves of astronomical transients",
    long_description=readme,
    author="Griffin Hosseinzadeh",
    author_email='griffin0@arizona.edu',
    url='https://github.com/griffin-h/lightcurve_fitting',
    python_requires='>={}'.format('.'.join(str(n) for n in min_version)),
    packages=find_packages(exclude=['docs', 'tests']),
    entry_points={
        'console_scripts': [
            # 'command = some.module:some_function',
        ],
    },
    include_package_data=True,
    package_data={
        'lightcurve_fitting': [
            'example/SN2016bkv.txt',
            'filters/*',
            'models/sifto.dat',
            'serif.mplstyle',
        ]
    },
    install_requires=requirements,
    license="GNU General Public License v3 (GPLv3)",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
    ],
)
